import io
import os
import tempfile
import typing
from typing import Any, Dict, Tuple

import pandas as pd
from pyjackson.decorators import type_field

from algolink.core.objects.artifacts import ArtifactCollection, LazyBlob
from algolink.core.objects.dataset_source import Dataset
from algolink.ext.pandas import DataFrameType
from algolink.ext.pandas.dataset import has_index, reset_index
from algolink.repository.dataset.artifact import DatasetReader, DatasetWriter

PANDAS_DATA_FILE = 'data.pd'


@type_field('type')
class PandasFormat:
    """ABC for reading and writing different formats supported in pandas

    :param read_args: additional arguments for reading
    :param write_args: additional arguments for writing
    """
    type: str = None
    read_func: typing.Callable = None
    write_func: typing.Callable = None
    buffer_type: typing.Type[typing.IO] = None

    def __init__(self, read_args: Dict[str, Any] = None, write_args: Dict[str, Any] = None):
        self.write_args = write_args or {}
        self.read_args = read_args or {}

    def read(self, file_or_path):
        """Read DataFrame

        :param file_or_path: source for read function"""
        kwargs = self.add_read_args()
        kwargs.update(self.read_args)
        return type(self).read_func(file_or_path, **kwargs)

    def write(self, dataframe) -> typing.IO:
        """Write DataFrame to buffer

        :param dataframe: DataFrame to write
        """
        buf = self.buffer_type()
        kwargs = self.add_write_args()
        kwargs.update(self.write_args)
        if has_index(dataframe):
            dataframe = reset_index(dataframe)
        type(self).write_func(dataframe, buf, **kwargs)
        return buf

    def add_read_args(self) -> Dict[str, Any]:
        """Fuction with additional read argumnets for child classes to override"""
        return {}

    def add_write_args(self) -> Dict[str, Any]:
        """Fuction with additional write argumnets for child classes to override"""
        return {}


class PandasFormatCsv(PandasFormat):
    type = 'csv'
    read_func = pd.read_csv
    write_func = pd.DataFrame.to_csv
    buffer_type = io.StringIO

    def add_write_args(self) -> Dict[str, Any]:
        return {'index': False}


class PandasFormatJson(PandasFormat):
    type = 'json'
    read_func = pd.read_json
    write_func = pd.DataFrame.to_json
    buffer_type = io.StringIO

    def add_write_args(self) -> Dict[str, Any]:
        return {'date_format': 'iso', 'date_unit': 'ns'}

    def read(self, file_or_path):
        # read_json creates index for some reason
        return super(PandasFormatJson, self).read(file_or_path).reset_index(drop=True)


class PandasFormatHtml(PandasFormat):
    type = 'html'
    read_func = pd.read_html
    write_func = pd.DataFrame.to_html
    buffer_type = io.StringIO

    def add_write_args(self) -> Dict[str, Any]:
        return {'index': False}

    def read(self, file_or_path):
        # read_html returns list of dataframes
        df = super(PandasFormatHtml, self).read(file_or_path)
        return df[0]


class PandasFormatExcel(PandasFormat):
    type = 'excel'
    read_func = pd.read_excel
    write_func = pd.DataFrame.to_excel
    buffer_type = io.BytesIO

    def add_write_args(self) -> Dict[str, Any]:
        return {'index': False}


class PandasFormatHdf(PandasFormat):
    type = 'hdf'
    read_func = pd.read_hdf
    write_func = pd.DataFrame.to_hdf
    buffer_type = io.BytesIO

    key = 'data'

    def add_write_args(self) -> Dict[str, Any]:
        return {'key': self.key}

    def write(self, dataframe) -> typing.IO:
        # to_hdf can write only to file or HDFStore, so there's that
        kwargs = self.add_write_args()
        kwargs.update(self.write_args)
        if has_index(dataframe):
            dataframe = reset_index(dataframe)
        path = tempfile.mktemp(suffix='.hd5', dir='.')  # tempfile.TemporaryDirectory breaks on windows for some reason
        try:
            type(self).write_func(dataframe, path, **kwargs)
            with open(path, 'rb') as f:
                return self.buffer_type(f.read())
        finally:
            os.unlink(path)

    def add_read_args(self) -> Dict[str, Any]:
        return {'key': self.key}

    def read(self, file_or_path):
        if not isinstance(file_or_path, str):
            path = tempfile.mktemp('.hd5', dir='.')
            try:
                with open(path, 'wb') as f:
                    f.write(file_or_path.read())
                df = super().read(path)
            finally:
                os.unlink(path)
        else:
            df = super().read(file_or_path)

        return df.reset_index(drop=True)


class PandasFormatFeather(PandasFormat):
    type = 'feather'
    read_func = pd.read_feather
    write_func = pd.DataFrame.to_feather
    buffer_type = io.BytesIO


class PandasFormatParquet(PandasFormat):
    type = 'parquet'
    read_func = pd.read_parquet
    write_func = pd.DataFrame.to_parquet
    buffer_type = io.BytesIO


# class PandasFormatStata(PandasFormat): # TODO int32 converts to int64 for some reason
#     type = 'stata'
#     read_func = pd.read_stata
#     write_func = pd.DataFrame.to_stata
#     buffer_type = io.BytesIO
#
#     def add_write_args(self) -> Dict[str, Any]:
#         return {'write_index': False}
#
# class PandasFormatPickle(PandasFormat): # TODO buffer closed error for some reason
#     type = 'pickle'
#     read_func = pd.read_pickle
#     write_func = pd.DataFrame.to_pickle
#     buffer_type = io.BytesIO


PANDAS_FORMATS = {f.type: f() for f in PandasFormat._subtypes.values() if f is not PandasFormat}


class PandasReader(DatasetReader):
    """DatasetReader for pandas dataframes

    :param format: PandasFormat instance to use
    :param data_type: DataFrameType to use for aliging read data
    """

    def __init__(self, format: PandasFormat, data_type: DataFrameType):
        self.data_type = data_type
        self.format = format

    def read(self, artifacts: ArtifactCollection) -> Dataset:
        with artifacts.blob_dict() as blobs, blobs[PANDAS_DATA_FILE].bytestream() as b:
            return Dataset.from_object(self.data_type.align(self.format.read(b)))


class PandasWriter(DatasetWriter):
    """DatasetWriter for pandas dataframes

    :param format: PandasFormat instance to use
    """

    def __init__(self, format: PandasFormat):
        self.format = format

    def write(self, dataset: Dataset) -> Tuple[DatasetReader, ArtifactCollection]:
        blob = LazyBlob(lambda: self.format.write(dataset.data))
        return PandasReader(self.format, dataset.dataset_type), ArtifactCollection.from_blobs({PANDAS_DATA_FILE: blob})



from sklearn.datasets import load_diabetes
from sklearn.linear_model import LinearRegression

X, y = load_diabetes(True)

lr = LinearRegression()
lr.fit(X, y)
import algolink
model = algolink.create_model(lr, X, model_name='diabetes_model_11')
model



import base64
import os
import tempfile
import zlib
from abc import abstractmethod
from typing import Any, Dict

from pyjackson.decorators import cached_property, type_field

from .base import AlgoLinkParams
from .requirements import Requirements
from .wrapper import PickleModelIO
from algolink.utils.importing import import_string
from algolink.utils.module import get_object_requirements


@type_field('type')
class Metric(AlgoLinkParams):
    @abstractmethod
    def evaluate(self, truth, prediction):
        raise NotImplementedError()


class LibFunctionMetric(Metric):
    def __init__(self, function: str, args: Dict[str, Any] = None, invert_input: bool = False):
        self.invert_input = invert_input
        self.args = args or {}
        self.function = function

    @cached_property
    def _function(self):
        return import_string(self.function)

    def evaluate(self, truth, prediction):
        if self.invert_input:
            return self._function(prediction, truth, **self.args)
        else:
            return self._function(truth, prediction, **self.args)


class CallableMetricWrapper:
    def __init__(self, artifacts: Dict[str, str], requirements: Requirements):
        self.artifacts = artifacts
        self.requirements = requirements
        self.callable = None

    def bind(self, callable):
        self.callable = callable
        return self

    @staticmethod
    def compress(s: bytes) -> str:
        """
        Helper method to compress source code

        :param s: source code
        :return: base64 encoded string of zipped source
        """
        zp = zlib.compress(s)
        b64 = base64.standard_b64encode(zp)
        return b64.decode('utf8')

    @staticmethod
    def decompress(s: str) -> bytes:
        """
        Helper method to decompress source code

        :param s: compressed source code
        :return: decompressed source code
        """
        zp = base64.standard_b64decode(s.encode('utf8'))
        src = zlib.decompress(zp)
        return src

    @classmethod
    def from_callable(cls, callable):
        reqs = get_object_requirements(callable)
        with PickleModelIO().dump(callable) as artifacts:
            payload = {path: cls.compress(bts) for path, bts in artifacts.bytes_dict().items()}
        return CallableMetricWrapper(payload, reqs).bind(callable)

    def load(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            for path, art in self.artifacts.items():
                with open(os.path.join(tmpdir, path), 'wb') as f:
                    f.write(self.decompress(art))
            self.callable = PickleModelIO().load(tmpdir)


class CallableMetric(Metric):
    def __init__(self, wrapper: CallableMetricWrapper):
        self.wrapper = wrapper

    def evaluate(self, truth, prediction):
        if self.wrapper.callable is None:
            self.wrapper.load()
        return self.wrapper.callable(truth, prediction)



from abc import abstractmethod

from algolink.core.objects.dataset_source import Dataset, DatasetSource


class DatasetRepository:
    """Base class for persisting datasets"""

    @abstractmethod
    def save(self, dataset_id: str, dataset: Dataset) -> DatasetSource:
        """Method to save dataset to this repository

        :param dataset_id: string identifier
        :param dataset: dataset to save
        :returns: DatasetSource that produces same Dataset"""

    @abstractmethod
    def delete(self, dataset_id: str):
        """Method to delete dataset from this repository

        :param dataset_id: dataset identifier
        """



import ast
import inspect
import io
import os
import re
import sys
import threading
import warnings
from collections import namedtuple
from functools import wraps
from pickle import PickleError
from types import FunctionType, LambdaType, MethodType, ModuleType

import requests
from isort.finders import FindersManager
from isort.settings import default

from ..core.objects.requirements import (MODULE_PACKAGE_MAPPING, CustomRequirement, InstallableRequirement,
                                               Requirements)
from . import importing
from .importing import import_module
from .log import logger
from .pickling import AlgoLinkPickler

PYTHON_BASE = os.path.dirname(threading.__file__)


def analyze_module_imports(module_path):
    module = importing.import_module(module_path)
    requirements = set()
    for name, obj in module.__dict__.items():
        if isinstance(obj, ModuleType):
            mod = obj
        else:
            mod = get_object_base_module(obj)
        if is_installable_module(mod) and not is_private_module(mod):
            requirements.add(get_module_repr(mod))

    return requirements


def check_pypi_module(module_name, module_version=None, raise_on_error=False, warn_on_error=True):
    """
    Checks that module with given name and (optionally) version exists in PyPi repository.

    :param module_name: name of module to look for in PyPi
    :param module_version: (optional) version of module to look for in PyPi
    :param raise_on_error: raise `ValueError` if module is not found in PyPi instead of returning `False`
    :param warn_on_error: print a warning if module is not found in PyPi
    :return: `True` if module found in PyPi, `False` otherwise
    """
    r = requests.get('https://pypi.org/pypi/{}/json'.format(module_name))
    if r.status_code != 200:
        msg = 'Cant find package {} in PyPi'.format(module_name)
        if raise_on_error:
            raise ValueError(msg)
        elif warn_on_error:
            warnings.warn(msg)
        return False
    if module_version is not None and module_version not in r.json()['releases']:
        msg = 'Cant find package version {}=={} in PyPi'.format(module_name, module_version)
        if raise_on_error:
            raise ImportError(msg)
        elif warn_on_error:
            warnings.warn(msg)
        return False
    return True


def get_object_base_module(obj: object) -> ModuleType:
    """
    Determines base module of module given object comes from.

    >>> import numpy
    >>> get_object_base_module(numpy.random.Generator)
    <module 'numpy' from '...'>

    Essentially this function is a combination of :func:`get_object_module` and :func:`get_base_module`.

    :param obj: object to determine base module for
    :return: Python module object for base module
    """
    mod = inspect.getmodule(obj)
    return get_base_module(mod)


def get_base_module(mod: ModuleType):
    """
    Determines base module for given module.

    >>> import numpy
    >>> get_base_module(numpy.random)
    <module 'numpy' from '...'>

    :param mod: Python module object to determine base module for
    :return: Python module object for base module
    """
    if mod is None:
        mod = inspect.getmodule(type(mod))
    if mod is None:
        return None
    base, _sep, _stem = mod.__name__.partition('.')
    return sys.modules[base]


def get_object_module(obj: object) -> ModuleType:
    """
    Determines module given object comes from

    >>> import numpy
    >>> get_object_module(numpy.ndarray)
    <module 'numpy' from '...'>

    :param obj: obj to determine module it comes from
    :return: Python module object for object module
    """
    return inspect.getmodule(obj)


def _create_section(section):
    def is_section(cls: 'ISortModuleFinder', module: str):
        cls.init()
        if module in cls.instance.module2section:
            mod_section = cls.instance.module2section[module]
        else:
            mod_section = cls.instance.finder.find(module)
            cls.instance.module2section[module] = mod_section
        return mod_section == section

    return is_section


class ISortModuleFinder:
    """
    Determines type of module: standard library (:meth:`ISortModuleFinder.is_stdlib`) or
    third party (:meth:`ISortModuleFinder.is_thirdparty`).
    This class uses `isort` library heuristics with some modifications.
    """
    instance: 'ISortModuleFinder' = None

    def __init__(self):
        config = default.copy()
        config['known_first_party'].append('algolink')
        config['known_third_party'].append('xgboost')
        config['known_standard_library'].extend(
            ['opcode', 'nturl2path',  # pytest requirements missed by isort
             'pkg_resources',  # EBNT-112: workaround for imports from setup.py (see build/builder/docker.py)
             'posixpath', 'setuptools',
             'pydevconsole', 'pydevd_tracing', 'pydev_ipython.matplotlibtools', 'pydev_console.protocol',
             'pydevd_file_utils', 'pydevd_plugins.extensions.types.pydevd_plugins_django_form_str', 'pydev_console',
             'pydev_ipython', 'pydevd_plugins.extensions.types.pydevd_plugin_numpy_types',
             'pydevd_plugins.extensions.types.pydevd_helpers', 'pydevd_plugins', 'pydevd_plugins.extensions.types',
             'pydevd_plugins.extensions', 'pydev_ipython.inputhook'])  # "built-in" pydev (and pycharm) modules
        section_names = config['sections']
        sections = namedtuple('Sections', section_names)(*[name for name in section_names])
        self.finder = FindersManager(config, sections)
        self.module2section = {}

    @classmethod
    def init(cls):
        if cls.instance is None:
            cls.instance = cls()

    is_stdlib = classmethod(_create_section('STDLIB'))
    is_thirdparty = classmethod(_create_section('THIRDPARTY'))


def is_private_module(mod: ModuleType):
    """
    Determines that given module object represents private module.

    :param mod: module object to use
    :return: boolean flag
    """
    return mod.__name__.startswith('_')


def is_pseudo_module(mod: ModuleType):
    """
    Determines that given module object represents pseudo (aka Python "magic") module.

    :param mod: module object to use
    :return: boolean flag
    """
    return mod.__name__.startswith('__') and mod.__name__.endswith('__')


def is_extension_module(mod: ModuleType):
    """
    Determines that given module object represents native code extension module.

    :param mod: module object to use
    :return: boolean flag
    """
    try:
        path = mod.__file__
        return any(path.endswith(ext) for ext in {'.so', '.pyd'})
    except AttributeError:
        return True


def is_installable_module(mod: ModuleType):
    """
    Determines that given module object represents PyPi-installable (aka third party) module.

    :param mod: module object to use
    :return: boolean flag
    """
    return ISortModuleFinder.is_thirdparty(mod.__name__)
    # return hasattr(mod, '__file__') and mod.__file__.startswith(PYTHON_BASE) and 'site-packages' in mod.__file__


def is_builtin_module(mod: ModuleType):
    """
    Determines that given module object represents standard library (aka builtin) module.

    :param mod: module object to use
    :return: boolean flag
    """
    return ISortModuleFinder.is_stdlib(mod.__name__)


def is_algolink_module(mod: ModuleType):
    """
    Determines that given module object is algolink module

    :param mod: module object to use
    :return: boolean flag
    """
    return mod.__name__ == 'algolink' or mod.__name__.startswith('algolink.')


def is_local_module(mod: ModuleType):
    """
    Determines that given module object represents local module.
    Local module is a module (Python file) which is not from standard library and not installed via pip.

    :param mod: module object to use
    :return: boolean flag
    """
    return (not is_pseudo_module(mod) and not is_algolink_module(mod) and not is_builtin_module(mod) and
            not is_installable_module(mod) and not is_extension_module(mod))


def is_from_installable_module(obj: object):
    """
    Determines that given object comes from PyPi-installable (aka third party) module.

    :param obj: object to check
    :return: boolean flag
    """
    return is_installable_module(get_object_base_module(obj))


def get_module_version(mod: ModuleType):
    """
    Determines version of given module object.

    :param mod: module object to use
    :return: version as `str` or `None` if version could not be determined
    """
    try:
        return mod.__version__
    except AttributeError:
        for name in os.listdir(os.path.dirname(mod.__file__)):
            m = re.match(re.escape(mod.__name__) + '-(.+)\\.dist-info', name)
            if m:
                return m.group(1)
        return None


def get_python_version():
    """
    :return: Current python version in 'major.minor.micro' format
    """
    major, minor, micro, *_ = sys.version_info
    return f'{major}.{minor}.{micro}'


def get_package_name(mod: ModuleType) -> str:
    """
    Determines PyPi package name for given module object

    :param mod: module object to use
    :return: name as `str`
    """
    if mod is None:
        raise ValueError('mod must not be None')
    name = mod.__name__
    return MODULE_PACKAGE_MAPPING.get(name, name)


def get_module_repr(mod: ModuleType, validate_pypi=False) -> str:
    """
    Builds PyPi `requirements.txt`-compatible representation of given module object

    :param mod: module object to use
    :param validate_pypi: if `True` (default is `False`) perform representation validation in PyPi repository
    :return: representation as `str`
    """
    if mod is None:
        raise ValueError('mod must not be None')
    mod_name = get_package_name(mod)
    mod_version = get_module_version(mod)
    rpr = '{}=={}'.format(mod_name, mod_version)
    if validate_pypi:
        check_pypi_module(mod_name, mod_version, raise_on_error=True)
    return rpr


def get_module_as_requirement(mod: ModuleType, validate_pypi=False) -> InstallableRequirement:
    """
    Builds AlgoLink representation of given module object

    :param mod: module object to use
    :param validate_pypi: if `True` (default is `False`) perform representation validation in PyPi repository
    :return: representation as :class:`.InstallableRequirement`
    """
    mod_version = get_module_version(mod)
    if validate_pypi:
        mod_name = get_package_name(mod)
        check_pypi_module(mod_name, mod_version, raise_on_error=True)
    return InstallableRequirement(mod.__name__, mod_version)


def get_local_module_reqs(mod):
    tree = ast.parse(inspect.getsource(mod))
    imports = []
    for statement in tree.body:
        if isinstance(statement, ast.Import):
            imports += [(n.name, None) for n in statement.names]
        elif isinstance(statement, ast.ImportFrom):
            if statement.level == 0:
                imp = (statement.module, None)
            else:
                imp = ('.' + statement.module, mod.__package__)
            imports.append(imp)

    result = [import_module(i, p) for i, p in imports]
    if mod.__file__.endswith('__init__.py'):
        # add loaded subpackages
        prefix = mod.__name__ + '.'
        result += [mod for name, mod in sys.modules.items() if name.startswith(prefix)]
    return result


def add_closure_inspection(f):
    @wraps(f)
    def wrapper(pickler: '_AlgoLinkRequirementAnalyzer', obj):
        closure = inspect.getclosurevars(obj)
        for field in ['nonlocals', 'globals']:
            for o in getattr(closure, field).values():
                if isinstance(o, ModuleType):
                    pickler._add_requirement(o)
                else:
                    pickler.save(o)

        if is_from_installable_module(obj):
            return f(pickler, obj)

        # to add from local imports inside user (non PIP package) code
        tree = ast.parse(inspect.getsource(obj).strip())

        class ImportFromVisitor(ast.NodeVisitor):
            def visit_ImportFrom(self, node: ast.ImportFrom):  # noqa
                warnings.warn(f'Detected local import in {obj.__module__}.{obj.__name__}')
                if node.level == 0:
                    mod = import_module(node.module)
                else:
                    mod = import_module('.' + node.module, get_object_module(obj).__package__)
                pickler._add_requirement(mod)

        ImportFromVisitor().visit(tree)

        return f(pickler, obj)

    return wrapper


class _AlgoLinkRequirementAnalyzer(AlgoLinkPickler):
    ignoring = (
        'dill',
        'algolink',
        'tests'  # pytest scans all test modules and all their imports are treated as requirements
    )
    dispatch = AlgoLinkPickler.dispatch.copy()

    add_closure_for = [FunctionType, MethodType, staticmethod, classmethod, LambdaType]
    dispatch.update({
        t: add_closure_inspection(AlgoLinkPickler.dispatch[t]) for t in add_closure_for
    })

    def __init__(self, *args, **kwargs):
        super().__init__(io.BytesIO(), *args, **kwargs)  # TODO maybe patch memo and other stuff too
        self.framer.write = self.skip_write
        self.write = self.skip_write
        self.memoize = self.skip_write
        self.seen = set()
        self._modules = set()

    @property
    def custom_modules(self):
        return set(m for m in self._modules if not is_installable_module(m))

    def to_requirements(self):
        r = Requirements()

        for mod in self._modules:
            if is_installable_module(mod):
                r.add(get_module_as_requirement(get_base_module(mod)))
            elif is_local_module(mod):
                r.add(CustomRequirement.from_module(mod))
        return r

    def _should_ignore(self, mod: ModuleType):
        return any(mod.__name__.startswith(i) for i in self.ignoring) or \
               is_private_module(mod) or is_pseudo_module(mod)

    def _add_requirement(self, obj_or_module):
        if not isinstance(obj_or_module, ModuleType):
            try:
                module = get_object_module(obj_or_module)
            except AttributeError as e:
                # Some internal Tensorflow 2.x object crashes `inspect` module on Python 3.6
                logger.debug('Skipping dependency analysis for %s because of %s: %s', obj_or_module,
                             type(e).__name__, e)
                return
        else:
            module = obj_or_module

        if module is not None and not self._should_ignore(module):
            self._modules.add(module)
            if is_local_module(module):
                # add imports of this module
                for local_req in get_local_module_reqs(module):
                    if local_req in self._modules:
                        continue
                    self._add_requirement(local_req)

    def save(self, obj, save_persistent_id=True):
        if id(obj) in self.seen:
            return
        self.seen.add(id(obj))
        self._add_requirement(obj)
        try:
            return super(AlgoLinkPickler, self).save(obj, save_persistent_id)
        except (ValueError, TypeError, PickleError) as e:
            # if object cannot be serialized, it's probably a C object and we don't need to go deeper
            logger.debug('Skipping dependency analysis for %s because of %s: %s', obj, type(e).__name__, e)

    def skip_write(self, *args, **kwargs):
        pass


def get_object_requirements(obj) -> Requirements:
    """
    Analyzes packages required for given object to perform its function.
    This function uses `pickle`/`dill` libraries serialization hooks internally.
    Thus result of this function depend on given object being serializable by `pickle`/`dill` libraries:
    all nodes in objects graph which can't be serialized are skipped and their dependencies are lost.

    :param obj: obj to analyze
    :return: :class:`.Requirements` object containing all required packages
    """
    a = _AlgoLinkRequirementAnalyzer(recurse=True)
    a.dump(obj)
    return a.to_requirements()



from abc import abstractmethod
from typing import Dict, List

from pyjackson import deserialize
from pyjackson.errors import DeserializationError, SerializationError

from algolink.config import Config, Core, Param
from algolink.runtime.interface import ExecutionError, Interface, InterfaceLoader
from algolink.runtime.utils import registering_type
from algolink.utils.classproperty import classproperty
from algolink.utils.log import rlogger

_ServerBase = registering_type('server')


class Server(_ServerBase):
    """
    Base class for AlgoLink servers
    """
    additional_sources: List[str] = []
    additional_binaries: List[str] = []
    additional_envs: Dict[str, str] = {}
    additional_options: Dict[str, str] = {}

    @staticmethod
    def get(class_path) -> 'Server':
        """
        Gets a fresh instance of given server implementation

        :param class_path: full name of server implementation
        :return: server object
        """

        # importing.import_module(class_path)
        return _ServerBase.get(class_path)()

    @abstractmethod
    def run(self, executor: Interface):
        """
        Main server method which "executes" given interface. Should be implemented by subclasses.

        :param executor: interface to "execute"
        :return: nothing
        """

        pass  # pragma: no cover

    def start(self, loader: InterfaceLoader):
        """
        Starts server "execution" for given loader: loads an interface and "executes" it

        :param loader: loader to take interface from
        :return: nothing
        """

        interface = loader.load()
        rlogger.info('Running server %s', self)
        return self.run(interface)

    @classproperty
    def type(cls):
        return f'{cls.__module__}.{cls.__name__}'


class HTTPServerConfig(Config):
    host = Param('host', default='0.0.0.0', parser=str)
    port = Param('port', default='9000', parser=int)


if Core.DEBUG:
    HTTPServerConfig.log_params()


class MalformedHTTPRequestException(Exception):
    def __init__(self, message: str):
        self._message = message

    def code(self):
        return 400

    def response_body(self):
        return {'ok': False, 'error': self._message}


class BaseHTTPServer(Server):
    """
    HTTP-based AlgoLink runtime server.

    Interface definition is exposed for clients via HTTP GET call to `/interface.json`,
    method calls - via HTTP POST calls to `/<name>`,
    server health check - via HTTP GET call to `/health`.

    Host to which server binds is configured via `ALGOLINK_HOST` environment variable:
    default is `0.0.0.0` which means any local or remote, for rejecting remote connections use `localhost` instead.

    Port to which server binds to is configured via `ALGOLINK_PORT` environment variable: default is 9000.
    """

    @staticmethod
    def _deserialize_json(interface: Interface, method: str, request_json: dict):
        args = {a.name: a for a in interface.exposed_method_args(method)}
        try:
            return {k: deserialize(v, args[k].type) for k, v in request_json.items()}
        except KeyError:
            raise MalformedHTTPRequestException(
                f'Invalid request: arguments are {set(args.keys())}, got {set(request_json.keys())}')
        except DeserializationError as e:
            raise MalformedHTTPRequestException(e.args[0])

    @staticmethod
    def _execute_method(interface: Interface, method: str, request_data, algolink_id: str):
        rlogger.debug('Got request for [%s]: %s', algolink_id, request_data)

        try:
            result = interface.execute(method, request_data)
        except (ExecutionError, SerializationError) as e:
            raise MalformedHTTPRequestException(e.args[0])

        if isinstance(result, bytes):
            rlogger.debug('Got response for [%s]: <binary content>', algolink_id)
            return result

        rlogger.debug('Got response for [%s]: %s', result)
        return {'ok': True, 'data': result}



import contextlib
import io
import os
import shutil
import sys
import tempfile
import typing
from abc import abstractmethod
from copy import copy

from pyjackson.core import Unserializable
from pyjackson.decorators import make_string, type_field

from .base import AlgoLinkParams

StreamContextManager = typing.Iterable[typing.BinaryIO]


@type_field('type')
class Blob(AlgoLinkParams):
    """
    This class is a base class for blobs.
    Blob is a binary payload, which can be accessed either through
    :meth:`~Blob.bytestream` context manager, which returns file-like object, or through
    :meth:`~Blob.materialize` method, which places a file in local fs

    Must be pyjackson-able or marked Unserializable
    """
    type = None

    @abstractmethod
    def materialize(self, path):
        """
        Implementation must write blob's payload as a file to local fs

        :param path: path to write file
        """
        pass  # pragma: no cover

    @abstractmethod
    @contextlib.contextmanager
    def bytestream(self) -> StreamContextManager:
        """
        Implementation must be a context manager (using @contextlib.contextmanager decorator)
        that returns file-like object with blob's payload

        :yields: file-like object
        """
        pass  # pragma: no cover

    def bytes(self) -> bytes:
        """
        Returns blob's bytes

        :return: bytes
        """
        with self.bytestream() as bs:
            return bs.read()


class LocalFileBlob(Blob):
    """
    Blob implementation for local file

    :param path: path to local file
    """
    type = 'local_file'

    def __init__(self, path: str):
        self.path = path

    def materialize(self, path):
        """
        Copies local file to another path

        :param path: target path
        """
        os.makedirs(os.path.dirname(path), exist_ok=True)
        shutil.copy(self.path, path)

    @contextlib.contextmanager
    def bytestream(self) -> StreamContextManager:
        """
        Opens file for reading

        :return: file handler
        """
        with open(self.path, 'rb') as f:
            yield f


# noinspection PyAbstractClass
class MaterializeOnlyBlobMixin(Blob):
    """
    Mixin for blobs which always have to be materialized first
    """

    def _get_file_name(self):
        return 'blob'

    @contextlib.contextmanager
    def bytestream(self) -> StreamContextManager:
        """
        Materializes blob to temporary dir and returns it's file handler

        :return: file handler
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, self._get_file_name())
            self.materialize(path)
            with open(path, 'rb') as f:
                yield f


class InMemoryBlob(Blob, Unserializable):
    """
    Blob implementation for in-memory bytes

    :param payload: bytes
    """
    type = 'inmemory'

    def __init__(self, payload: bytes):
        self.payload = payload

    def materialize(self, path):
        """
        Writes payload to path

        :param path: target path
        """
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'wb') as f:
            f.write(self.payload)

    @contextlib.contextmanager
    def bytestream(self) -> StreamContextManager:
        """
        Creates BytesIO object from bytes

        :yields: file-like object
        """
        yield io.BytesIO(self.payload)


class LazyBlob(Blob, Unserializable):
    """Represents a lazy blob, which is computed only when needed

    :param source: function with no arguments, that must return str, bytes or file-like object
    :param encoding: encoding for payload if source returns str of io.StringIO
    """
    def __init__(self, source: typing.Callable[[], typing.Union[str, bytes, typing.IO]], encoding: str = 'utf8'):
        self.encoding = encoding
        self.source = source

    def materialize(self, path):
        """
        Writes payload to path

        :param path: target path
        """
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'wb') as f:
            source = self.source()
            if not isinstance(source, (str, bytes)):
                source.seek(0)
                source = source.read()
            if isinstance(source, str):
                source = source.encode(self.encoding)
            f.write(source)

    @contextlib.contextmanager
    def bytestream(self) -> StreamContextManager:
        """
        Creates BytesIO object from bytes

        :yields: file-like object
        """
        source = self.source()
        source.seek(0)
        yield source


@type_field('type')
class ArtifactCollection(AlgoLinkParams):
    """
    Base class for artifact collection.
    Artifact collection is a number of named artifacts, represented by Blob's

    Must be pyjackson-able
    """
    type = None

    @abstractmethod
    def materialize(self, path):
        """
        Implementation must write artifacts payload to local fs

        :param path: dir to write blobs
        """
        pass  # pragma: no cover

    @abstractmethod
    def bytes_dict(self) -> typing.Dict[str, bytes]:
        """
        Implementation must return a dict of artifact name -> artifact payload

        :returns: dict of artifact names -> artifact payloads
        """
        pass  # pragma: no cover

    @abstractmethod
    def blob_dict(self) -> typing.ContextManager[typing.Dict[str, Blob]]:
        """
        Implementation must be a context manager (using `@contextlib.contexmanager` decorator) that yields a dict of
        artifact name -> blob

        :yields: dict of artifact name -> blob
        """
        pass  # pragma: no cover

    def __add__(self, other):
        """
        Creates a :py:class:`CompositeArtifactCollection` with union of artifacts from self and other ArtifactCollection

        :param other:
        :return:
        """
        if not isinstance(other, ArtifactCollection):
            raise ValueError('Cant and {} to ArtifactCollection'.format(other))
        return CompositeArtifactCollection([self, other])

    @staticmethod
    def from_blobs(files: typing.Dict[str, 'Blob']):
        return Blobs(files)


@make_string
class Blobs(ArtifactCollection):
    """
    Artifact collection represented by a dictionary of blobs

    :param blobs: dict of name -> blob
    """
    type = 'blobs'

    def __init__(self, blobs: typing.Dict[str, Blob]):
        self.blobs = blobs

    def materialize(self, path):
        """
        Materializes artifacts to path

        :param path: target dir
        """
        os.makedirs(path, exist_ok=True)
        for name, blob in self.blobs.items():
            blob.materialize(os.path.join(path, name))

    def bytes_dict(self) -> typing.Dict[str, bytes]:
        return {name: b.bytes() for name, b in self.blobs.items()}

    @contextlib.contextmanager
    def blob_dict(self) -> typing.ContextManager[typing.Dict[str, Blob]]:
        """
        :yields: self.blobs
        """
        yield self.blobs


def _enter_all_cm(managers):
    entered = []
    dicts = []
    try:
        for m in managers:
            dicts.append(m.__enter__())
            entered.append(m)
    except Exception as e:
        for m in reversed(entered):
            try:
                m.__exit__(type(e), e, sys.exc_traceback)
            except Exception as e2:
                e = e2
        raise e
    return dicts


class _ExitAllCm:
    def __init__(self, managers):
        self.managers = managers

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        for m in reversed(self.managers):
            try:
                m.__exit__(exc_type, exc_val, exc_tb)
            except Exception as e:
                exc_type, exc_val, exc_tb = type(e), e, sys.exc_traceback


class CompositeArtifactCollection(ArtifactCollection):
    """
    Represents a merger of two or more ArtifactCollections

    :param artifacts: ArtifactCollections to merge
    """
    type = 'composite'

    def __init__(self, artifacts: typing.List[ArtifactCollection]):
        self.artifacts = artifacts

    def materialize(self, path):
        """
        Materializes every ArtifactCollection to path

        :param path: target dir
        """
        for a in self.artifacts:
            a.materialize(path)

    def bytes_dict(self) -> typing.Dict[str, bytes]:
        return {k: v for a in self.artifacts for k, v in a.bytes_dict().items()}

    @contextlib.contextmanager
    def blob_dict(self) -> typing.ContextManager[typing.Dict[str, Blob]]:
        """
        Enters all ArtifactCollections blob_dict context managers and returns their union

        :yields: name -> blob mapping
        """
        managers: typing.List[typing.ContextManager] = [a.blob_dict() for a in self.artifacts]
        dicts = _enter_all_cm(managers)
        with _ExitAllCm(managers):
            yield {name: blob for d in dicts for name, blob in d.items()}

    def __add__(self, other):
        if not isinstance(other, ArtifactCollection):
            raise ValueError('Cant and {} to ArtifactCollection'.format(other))
        arts = copy(self.artifacts)
        if isinstance(other, CompositeArtifactCollection):
            arts += other.artifacts
        else:
            arts.append(other)
        return CompositeArtifactCollection(arts)

    def __iadd__(self, other):
        if not isinstance(other, ArtifactCollection):
            raise ValueError('Cant and {} to ArtifactCollection')
        if isinstance(other, CompositeArtifactCollection):
            self.artifacts += other.artifacts
        else:
            self.artifacts.append(other)
        return self


class _RelativePathWrapper(ArtifactCollection):
    """
    ArtifactCollection that adds path as a parent dir to every artifact

    :param path: parent dir
    """
    type = 'relative'

    def __init__(self, artifact: ArtifactCollection, path: str):
        self.artifact = artifact
        self.path = path

    def materialize(self, path):
        return self.artifact.materialize(os.path.join(path, self.path))

    def bytes_dict(self) -> typing.Dict[str, bytes]:
        return {os.path.join(self.path, p): b for p, b in self.artifact.bytes_dict().items()}

    @contextlib.contextmanager
    def blob_dict(self) -> typing.Iterable[typing.Dict[str, Blob]]:
        with self.artifact.blob_dict() as blobs:
            yield {os.path.join(self.path, p): b for p, b in blobs.items()}



IS_ABSTRACT_METHOD_FIELD = '__isabstractmethod__'


def is_abstract_method(cls_or_method, method_name=None):
    """
    Checks that given method is abstract (has no body and should be implemented by subclass)

    :param cls_or_method: either a class in which method `method_name` is found or method itself
    :param method_name: unused if `cls_or_method` is a method or name of method to look in `cls_or_method` class for
    :return: boolean flag
    """
    if isinstance(cls_or_method, type):
        method = getattr(cls_or_method, method_name)
    else:
        method = cls_or_method

    return hasattr(method, IS_ABSTRACT_METHOD_FIELD) and getattr(method, IS_ABSTRACT_METHOD_FIELD)



import base64
import glob
import itertools
import json
import os
import zlib
from types import ModuleType
from typing import Dict, List, Type, TypeVar, Union

from pyjackson.decorators import make_string, type_field

from .base import AlgoLinkParams

# TODO i dont know how to do this better
MODULE_PACKAGE_MAPPING = {
    'sklearn': 'scikit-learn',
    'skimage': 'scikit-image'
}
PACKAGE_MODULE_MAPPING = {v: k for k, v in MODULE_PACKAGE_MAPPING.items()}


def read(path, bin=False):
    with open(path, 'r' if not bin else 'rb') as f:
        return f.read()


@type_field('type')
class Requirement(AlgoLinkParams):
    """
    Base class for python requirement
    """
    type = None


class PythonRequirement(Requirement):
    module = None


@make_string(include_name=True)
class InstallableRequirement(PythonRequirement):
    """
    This class represents pip-installable python library

    :param module: name of python module
    :param version: version of python package
    :param package_name: Optional. pip package name for this module, if it is different from module name
    """
    type = 'installable'

    def __init__(self, module: str, version: str = None, package_name: str = None):
        self.module = module
        self.version = version
        self.package_name = package_name

    @property
    def package(self):
        """
        Pip package name
        """
        return self.package_name or MODULE_PACKAGE_MAPPING.get(self.module, self.module)

    def to_str(self):
        """
        pip installable representation of this module
        """
        if self.version is not None:
            return '{}=={}'.format(self.package, self.version)
        return self.package

    @classmethod
    def from_module(cls, mod: ModuleType, package_name: str = None) -> 'InstallableRequirement':
        """
        Factory method to create :class:`InstallableRequirement` from module object

        :param mod: module object
        :param package_name: PIP package name if it is not equal to module name
        :return: :class:`InstallableRequirement`
        """
        from algolink.utils.module import get_module_version
        return InstallableRequirement(mod.__name__, get_module_version(mod), package_name)

    @classmethod
    def from_str(cls, name):
        """
        Factory method for creating :class:`InstallableRequirement` from string

        :param name: string representation
        :return: :class:`InstallableRequirement`
        """
        for rel in ['==', '>=', '<=']:  # TODO for now we interpret everything as exact version
            if rel in name:
                package, version = name.split(rel)
                return InstallableRequirement(package, version)

        return InstallableRequirement(name)  # FIXME for other relations like > < !=


@make_string(include_name=True)
class CustomRequirement(PythonRequirement):
    """
    This class represents local python code that you need as a requirement for your code

    :param name: filename of this code
    :param source64zip: zipped and base64-encoded source
    :param is_package: whether this code should be in %name%/__init__.py
    """
    type = 'custom'

    def __init__(self, name: str, source64zip: str, is_package: bool):
        self.source64zip = source64zip
        self.name = name
        self.is_package = is_package

    @staticmethod
    def from_module(mod: ModuleType) -> 'CustomRequirement':
        """
        Factory method to create :class:`CustomRequirement` from module object

        :param mod: module object
        :return: :class:`CustomRequirement`
        """
        is_package = mod.__file__.endswith('__init__.py')
        if is_package:
            pkg_dir = os.path.dirname(mod.__file__)
            par = os.path.dirname(pkg_dir)
            sources = {os.path.relpath(p, par): read(p, bin=True)
                       for p in glob.glob(os.path.join(pkg_dir, '**', '*'), recursive=True) if os.path.isfile(p)}
            src = CustomRequirement.compress_package(sources)
        else:
            src = CustomRequirement.compress(read(mod.__file__))
        return CustomRequirement(mod.__name__, src, is_package)

    @staticmethod
    def compress(s: str) -> str:
        """
        Helper method to compress source code

        :param s: source code
        :return: base64 encoded string of zipped source
        """
        zp = zlib.compress(s.encode('utf8'))
        b64 = base64.standard_b64encode(zp)
        return b64.decode('utf8')

    @staticmethod
    def compress_package(s: Dict[str, bytes]) -> str:
        sources = {
            path: base64.standard_b64encode(zlib.compress(payload)).decode('utf8') for path, payload in s.items()
        }
        return CustomRequirement.compress(json.dumps(sources))

    @staticmethod
    def decompress(s: str) -> str:
        """
        Helper method to decompress source code

        :param s: compressed source code
        :return: decompressed source code
        """
        zp = base64.standard_b64decode(s.encode('utf8'))
        src = zlib.decompress(zp)
        return src.decode('utf8')

    @staticmethod
    def decompress_package(s: str) -> Dict[str, bytes]:
        sources = json.loads(CustomRequirement.decompress(s))
        return {
            path: zlib.decompress(base64.standard_b64decode(payload.encode('utf8')))
            for path, payload in sources.items()
        }

    @property
    def module(self):
        """
        Module name for this requirement
        """
        return self.name.split('.')[0]

    @property
    def source(self) -> str:
        """
        Source code of this requirement
        """
        if not self.is_package:
            return CustomRequirement.decompress(self.source64zip)
        raise AttributeError("package requirement does not have source attribute")

    @property
    def sources(self) -> Dict[str, bytes]:
        if self.is_package:
            return CustomRequirement.decompress_package(self.source64zip)
        raise AttributeError("non package requirement does not have sources attribute")

    def to_sources_dict(self):
        """
        Mapping path -> source code for this requirement

        :return: dict path -> source
        """
        if self.is_package:
            return self.sources
        else:
            return {self.name.replace('.', '/') + '.py': self.source}


class FileRequirement(CustomRequirement):
    def __init__(self, name: str, source64zip: str):
        super().__init__(name, source64zip, False)

    def to_sources_dict(self):
        """
        Mapping path -> source code for this requirement

        :return: dict path -> source
        """
        return {self.name: self.source}

    @classmethod
    def from_path(cls, path: str):
        return FileRequirement(path, cls.compress(read(path)))


@make_string
class UnixPackageRequirement(Requirement):
    def __init__(self, package_name: str):
        self.package_name = package_name


T = TypeVar('T', bound=Requirement)


class Requirements(AlgoLinkParams):
    """
    A collection of requirements

    :param requirements: list of :class:`Requirement` instances
    """

    def __init__(self, requirements: List[Requirement] = None):
        self.requirements = requirements or []

    @property
    def installable(self) -> List[InstallableRequirement]:
        """
        List of installable requirements
        """
        return self.of_type(InstallableRequirement)

    @property
    def custom(self) -> List[CustomRequirement]:
        """
        List of custom requirements
        """
        return self.of_type(CustomRequirement)

    def of_type(self, type_: Type[T]) -> List[T]:
        """
        :param type_: type of requirements
        :return: List of requirements of type `type_`
        """
        return [r for r in self.requirements if isinstance(r, type_)]

    @property
    def modules(self) -> List[str]:
        """
        List of module names
        """
        return [r.module for r in self.of_type(PythonRequirement)]

    def add(self, requirement: Requirement):
        """
        Adds requirement to this collection

        :param requirement: :class:`Requirement` instance to add
        """
        if isinstance(requirement, InstallableRequirement):
            for r in self.installable:
                if r.package == requirement.package:
                    if r.version == requirement.version:
                        break
                    if r.version is not None and r.version != requirement.version:
                        raise ValueError('Conflicting versions for package {}: {} and {}'.format(r.package, r.version,
                                                                                                 requirement.version))
            else:
                self.requirements.append(requirement)
        elif isinstance(requirement, CustomRequirement):
            if requirement.is_package:
                for r in self.custom:
                    if r.name.startswith(requirement.name + '.') or r.name == requirement.name:
                        # existing req is subpackage or equals to new req
                        self.requirements.remove(r)
                    if requirement.name.startswith(r.name + '.'):
                        # new req is subpackage of existing
                        break
                else:
                    self.requirements.append(requirement)
            else:
                for r in self.custom:
                    if r.is_package and requirement.name.startswith(r.name + '.'):
                        # new req is from existing package
                        break
                    if not r.is_package and r.name == requirement.name:
                        # new req equals to existing
                        break
                else:
                    self.requirements.append(requirement)
        else:  # TODO better checks here
            if requirement not in self.requirements:
                self.requirements.append(requirement)

    def to_pip(self) -> List[str]:
        """
        :return: list of pip installable packages
        """
        return [r.to_str() for r in self.installable]

    def __add__(self, other: 'AnyRequirements'):
        other = resolve_requirements(other)
        res = Requirements([])
        for r in itertools.chain(self.requirements, other.requirements):
            res.add(r)
        return res

    def __iadd__(self, other: 'AnyRequirements'):
        return self + other


def resolve_requirements(other: 'AnyRequirements') -> Requirements:
    """
    Helper method to create :class:`Requirements` from any supported source.
    Supported formats: :class:`Requirements`, :class:`Requirement`, list of :class:`Requirement`,
    string representation or list of string representations

    :param other: requirement in supported format
    :return: :class:`Requirements` instance
    """
    if not isinstance(other, Requirements):
        if isinstance(other, list):
            if isinstance(other[0], str):
                other = Requirements([InstallableRequirement.from_str(r) for r in other])
            elif isinstance(other[0], Requirement):
                other = Requirements([r for r in other])
        elif isinstance(other, Requirement):
            other = Requirements([other])
        elif isinstance(other, str):
            other = Requirements([InstallableRequirement.from_str(other)])
        else:
            raise TypeError('only other Requirements, Requirement, list of Requirement objects, string '
                            '(or list of strings) can be added')
    return other


AnyRequirements = Union[Requirements, Requirement, List[Requirement], str, List[str]]



import inspect
from abc import ABC, abstractmethod
from types import ModuleType
from typing import List, Type

from algolink.utils.log import logger
from algolink.utils.module import get_object_base_module

ANALYZER_FIELD = '_analyzer'


class Hook(ABC):
    """
    Base class for Hooks
    """

    @abstractmethod
    def can_process(self, obj) -> bool:
        """
        Must return True if obj can be processed by this hook

        :param obj: object to analyze
        :return: True or False
        """
        pass  # pragma: no cover

    @abstractmethod
    def must_process(self, obj) -> bool:
        """
        Must return True if obj must be processed by this hook. "must" means you sure that no other hook should handle
        this object, for example this hook is for sklearn objects and obj is exactly that.

        :param obj: object to analyze
        :return: True or False
        """
        pass  # pragma: no cover

    @abstractmethod
    def process(self, obj, **kwargs):
        """
        Analyzes obj and returns result. Result type is determined by specific Hook class sub-hierarchy

        :param obj: object to analyze
        :param kwargs: additional information to be used for analysis
        :return: analysis result
        """
        pass  # pragma: no cover

    def __init_subclass__(cls, **kwargs):
        if hasattr(cls, '__init__'):
            init = getattr(cls, '__init__')
            arg_spec = inspect.getfullargspec(init)
            if len(arg_spec.args) > 1:
                raise ValueError('Hook type [{}] cannot have __init__ with arguments'.format(cls.__name__))

        if not inspect.isabstract(cls):
            for b in reversed(cls.__bases__):
                analyzer = getattr(b, ANALYZER_FIELD, None)
                if analyzer is not None:
                    analyzer.hooks.append(cls())
                    logger.debug('Registering %s to %s', cls.__name__, analyzer.__name__)
                    break
            else:
                raise ValueError(
                    '{} defines process method, but dont have any parents with attached Analyzer'.format(cls))
        super(Hook, cls).__init_subclass__(**kwargs)


# noinspection PyAbstractClass
class CanIsAMustHookMixin(Hook):
    """
    Mixin for cases when can_process equals to must_process
    """

    def can_process(self, obj) -> bool:
        """Returns same as :meth:`Hook.must_process`"""
        return self.must_process(obj)


# noinspection PyAbstractClass
class TypeHookMixin(CanIsAMustHookMixin):
    """
    Mixin for cases when hook must process objects of certain types
    """
    valid_types: List[Type] = None

    def must_process(self, obj) -> bool:
        """Returns True if obj is instance of one of valid types"""
        return any(isinstance(obj, t) for t in self.valid_types)


class BaseModuleHookMixin(CanIsAMustHookMixin, Hook):
    """
    Mixin for cases when hook must process all objects with certain base modules
    """

    @abstractmethod
    def is_valid_base_module_name(self, module_name: str) -> bool:
        """
        Must return True if module_name is valid for this hook

        :param module_name: module name
        :return: True or False
        """
        pass  # pragma: no cover

    def is_valid_base_module(self, base_module: ModuleType) -> bool:
        """
        Returns True if module is valid

        :param base_module: module object
        :return: True or False
        """
        if base_module is None:
            return False
        return self.is_valid_base_module_name(base_module.__name__)

    def must_process(self, obj):
        """Returns True if obj has valid base module"""
        return self.is_valid_base_module(get_object_base_module(obj))


class LibHookMixin(BaseModuleHookMixin):
    """
    Mixin for cases when hook must process all objects with certain base module
    """
    base_module_name = None

    def is_valid_base_module_name(self, base_module: str) -> bool:
        return base_module == self.base_module_name


def analyzer_class(hook_type: type, return_type: type):
    """
    Function to create separate hook hierarchies for analyzing different objects

    :param hook_type: Subtype of :py:class:`Hook`
    :param return_type: Type that this hierarchy will use as analysis result
    :return: Analyzer type
    """
    if hasattr(hook_type, ANALYZER_FIELD):
        raise ValueError('{} hook already have analyzer'.format(hook_type))

    class Analyzer:
        f"""
        Analyzer for {hook_type.__name__} hooks
        """
        hooks: List[hook_type] = []

        @classmethod
        def analyze(cls, obj, **kwargs) -> return_type:
            f"""
            Run {hook_type.__name__} hooks to analyze obj

            :param obj: objects to analyze
            :param kwargs: additional information to be used for analysis
            :return: Instance of {return_type.__name__}
            """
            return cls._find_hook(obj).process(obj, **kwargs)

        @classmethod
        def _find_hook(cls, obj) -> hook_type:
            hooks = []
            for hook in cls.hooks:
                if hook.must_process(obj):
                    logger.debug('processing class %s with %s', type(obj).__name__, hook.__class__.__name__)
                    return hook
                elif hook.can_process(obj):
                    hooks.append(hook)

            if not hooks:
                raise ValueError(
                    f'No suitable {hook_type.__name__} for object '
                    f'[{type(obj).__name__}] {obj}. Registered hooks: {cls.hooks}')
            elif len(hooks) > 1:
                raise ValueError(f'Multiple suitable hooks for object {obj} ({hooks})')

            return hooks[0]

    Analyzer.__name__ = '{}Analyzer'.format(hook_type.__name__)
    setattr(hook_type, ANALYZER_FIELD, Analyzer)
    return Analyzer



import contextlib
import os
import tempfile

import catboost
from catboost import CatBoostClassifier, CatBoostRegressor
from pyjackson.decorators import make_string

from algolink.core.analyzer import TypeHookMixin
from algolink.core.analyzer.model import BindingModelHook
from algolink.core.objects.artifacts import ArtifactCollection, Blobs, LocalFileBlob
from algolink.core.objects.wrapper import LibModelWrapperMixin, ModelIO, ModelWrapper


class CatBoostModelIO(ModelIO):
    """
    :class:`algolink.core.objects.ModelIO` for CatBoost models.
    """
    classifier_file_name = 'clf.cb'
    regressor_file_name = 'rgr.cb'

    @contextlib.contextmanager
    def dump(self, model) -> ArtifactCollection:
        """
        Dumps `catboost.CatBoostClassifier` or `catboost.CatBoostRegressor` instance to :class:`.LocalFileBlob` and
        creates :class:`.ArtifactCollection` from it

        :return: context manager with :class:`~algolink.core.objects.ArtifactCollection`
        """
        model_file = tempfile.mktemp()
        try:
            model.save_model(model_file)
            yield Blobs({self._get_model_file_name(model): LocalFileBlob(model_file)})
        finally:
            os.remove(model_file)

    def _get_model_file_name(self, model):
        if isinstance(model, CatBoostClassifier):
            return self.classifier_file_name
        return self.regressor_file_name

    def load(self, path):
        """
        Loads `catboost.CatBoostClassifier` or `catboost.CatBoostRegressor` instance from path

        :param path: path to load from
        """
        if os.path.exists(os.path.join(path, self.classifier_file_name)):
            model_type = CatBoostClassifier
        else:
            model_type = CatBoostRegressor

        model = model_type()
        model.load_model(os.path.join(path, self._get_model_file_name(model)))
        return model


class CatBoostModelWrapper(LibModelWrapperMixin):
    """
    :class:`algolink.core.objects.ModelWrapper` for CatBoost models.
    `.model` attribute is a `catboost.CatBoostClassifier` or `catboost.CatBoostRegressor` instance
    """
    libraries = [catboost]

    def __init__(self):
        super().__init__(CatBoostModelIO())

    def _exposed_methods_mapping(self):
        ret = {
            'predict': 'predict'
        }
        if isinstance(self.model, CatBoostClassifier):
            ret['predict_proba'] = 'predict_proba'
        return ret


@make_string(include_name=True)
class CatBoostModelHook(BindingModelHook, TypeHookMixin):
    """
    Hook for CatBoost models
    """
    valid_types = [CatBoostClassifier,  CatBoostRegressor]

    def _wrapper_factory(self) -> ModelWrapper:
        """
        Creates :class:`CatBoostModelWrapper` for CatBoost model object

        :return: :class:`CatBoostModelWrapper` instance
        """
        return CatBoostModelWrapper()



from .builder import BuilderBase, PythonBuildContext
from .runner import RunnerBase
from .provider import PipelineProvider, MLModelMultiProvider, MLModelProvider

__all__ = ['RunnerBase', 'BuilderBase', 'PythonBuildContext',
           'PipelineProvider', 'MLModelMultiProvider', 'MLModelProvider']



from abc import abstractmethod

from .base import Hook, analyzer_class
from ..objects.dataset_type import (PRIMITIVES, BytesDatasetType, DatasetType, DictDatasetType,
                                               ListDatasetType, PrimitiveDatasetType, TupleDatasetType,
                                               TupleLikeListDatasetType)


class DatasetHook(Hook):
    """
    Base hook type for :class:`DatasetAnalyzer`.
    Analysis result is an instance of :class:`~algolink.core.objects.DatasetType`
    """
    @abstractmethod
    def process(self, obj, **kwargs) -> DatasetType:
        """
        Analyzes obj and returns result. Result type is determined by specific Hook class sub-hierarchy

        :param obj: object to analyze
        :param kwargs: additional information to be used for analysis
        :return: analysis result
        """
        pass  # pragma: no cover


DatasetAnalyzer = analyzer_class(DatasetHook, DatasetType)


class PrimitivesHook(DatasetHook):
    """
    Hook for primitive data, for example when you model outputs just one int
    """
    def can_process(self, obj):
        if type(obj) in PRIMITIVES:
            return True

    def must_process(self, obj):
        return False

    def process(self, obj, **kwargs) -> DatasetType:
        return PrimitiveDatasetType(type(obj).__name__)


class OrderedCollectionHookDelegator(DatasetHook):
    """
    Hook for list/tuple data
    """
    def can_process(self, obj) -> bool:
        return isinstance(obj, (list, tuple))

    def must_process(self, obj) -> bool:
        return False

    def process(self, obj, **kwargs) -> DatasetType:
        if isinstance(obj, tuple):
            return TupleDatasetType([DatasetAnalyzer.analyze(o) for o in obj])

        py_types = {type(o) for o in obj}
        if len(obj) <= 1 or len(py_types) > 1:
            return TupleLikeListDatasetType([DatasetAnalyzer.analyze(o) for o in obj])

        if not py_types.intersection(PRIMITIVES):  # py_types is guaranteed to be singleton set here
            return TupleLikeListDatasetType([DatasetAnalyzer.analyze(o) for o in obj])

        # optimization for large lists of same primitive type elements
        return ListDatasetType(DatasetAnalyzer.analyze(obj[0]), len(obj))


class DictHookDelegator(DatasetHook):
    """
    Hook for dict data
    """
    def can_process(self, obj) -> bool:
        return isinstance(obj, dict)

    def must_process(self, obj) -> bool:
        return False

    def process(self, obj, **kwargs) -> DatasetType:
        try:
            items = {k: DatasetAnalyzer.analyze(o) for k, o in obj.items()}
        except ValueError:
            raise ValueError(f"Cant process {obj} with DictHookDelegator")
        return DictDatasetType(items)


class BytesDatasetHook(DatasetHook):
    """
    Hook for bytes objects
    """
    def process(self, obj, **kwargs) -> DatasetType:
        return BytesDatasetType()

    def can_process(self, obj) -> bool:
        return isinstance(obj, bytes)

    def must_process(self, obj) -> bool:
        return False



import typing
from abc import abstractmethod

from pyjackson.decorators import type_field

from algolink.core.objects import core
from algolink.core.objects.artifacts import ArtifactCollection, Blob


@type_field('type')
class ArtifactRepository:
    """
    Base abstract class for persistent repositories of artifacts
    """

    type = None
    MODEL_TYPE = 'model'

    def get_model_id(self, model: 'core.Model') -> str:
        model_id = model.id
        if model_id is None:
            raise ValueError('model_id cannot be "None"')
        return str(model_id)

    def push_model_artifacts(self, model: 'core.Model'):
        """
        Helper method which handles the most common model artifacts workflow.
        Based on :meth:`.ArtifactRepository.push_artifact` and :meth:`.core.Model.persist_artifacts` methods.

        :param model: model to store artifacts in the repository for
        :return: nothing
        """

        def _persister(artifact):
            with artifact.blob_dict() as files:
                return self.push_model_artifact(model, files)

        model.persist_artifacts(_persister)
        model.bind_artifact_repo(self)

    def push_model_artifact(self, model: 'core.Model', blobs: typing.Dict[str, Blob]) -> ArtifactCollection:
        """
        Stores given :class:`.Blob` artifacts in the repository and associates them with given model

        :param model: model to associate artifacts with
        :param blobs: artifacts to store in the repository
        :return: :class:`.ArtifactCollection` object containing stored artifacts
        :exception: :exc:`.ArtifactExistsError` if there are already artifacts stored for this model
        """
        model.bind_artifact_repo(self)
        return self.push_artifact(self.MODEL_TYPE, self.get_model_id(model), blobs)

    def get_model_artifact(self, model: 'core.Model') -> ArtifactCollection:
        """
        Gets artifacts for given model

        :param model: model to get artifacts for
        :return: :class:`.ArtifactCollection` object containing stored artifacts
        :exception: :exc:`.NoSuchArtifactError` if no artifacts were associated with given model
        """
        return self.get_artifact(self.MODEL_TYPE, self.get_model_id(model))

    def delete_model_artifact(self, model: 'core.Model'):
        """
        Deletes artifacts for given model

        :param model: model to delete artifacts for
        :return: nothing
        :exception: :exc:`.NoSuchArtifactError` if no artifacts were associated with given model
        """
        self.delete_artifact(self.MODEL_TYPE, self.get_model_id(model))
        model.unbind_artifact_repo()

    @abstractmethod
    def push_artifact(self, artifact_type: str, artifact_id: str, blobs: typing.Dict[str, Blob]) -> ArtifactCollection:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def get_artifact(self, artifact_type: str, artifact_id: str) -> ArtifactCollection:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def delete_artifact(self, artifact_type: str, artifact_id: str):
        raise NotImplementedError  # pragma: no cover


# noinspection PyAbstractClass
class RepoArtifactBlob(Blob):
    def __init__(self, repository: ArtifactRepository):
        self.repository = repository



from .dataset import DMatrixDatasetType, DMatrixHook
from .model import XGBoostModelHook, XGBoostModelWrapper
from .requirement import XgboostRequirementHook

__all__ = ['DMatrixHook', 'DMatrixDatasetType', 'XGBoostModelHook', 'XGBoostModelWrapper', 'XgboostRequirementHook']



from abc import abstractmethod
from typing import Callable, Dict, List

from pyjackson import deserialize, serialize
from pyjackson.core import Comparable, Field, Signature
from pyjackson.utils import get_function_signature

from algolink.core import objects
from algolink.runtime.utils import registering_type


def expose(class_method):
    """
    Decorator which exposes given method into interface

    :param class_method: method to expose
    :return: given method with modifications
    """

    class_method.is_exposed = True
    return class_method


class ExecutionError(Exception):
    """
    Exception which is raised when interface method is executed with arguments incompatible to its signature
    """

    pass


class InterfaceMetaclass(type):
    """
    Metaclass for :class:`Interface` which keeps track of exposed methods
    """

    def __new__(mcs, *args, **kwargs):
        new_cls = super().__new__(mcs, *args, **kwargs)
        new_cls.exposed = dict()
        for name, attr in new_cls.__dict__.items():
            if hasattr(attr, 'is_exposed') and getattr(attr, 'is_exposed'):
                new_cls.exposed[name] = get_function_signature(attr)
        return new_cls


class Interface(metaclass=InterfaceMetaclass):
    """
    Collection of executable methods with explicitly defined signatures
    """

    exposed: Dict[str, Signature] = {}
    executors: Dict[str, Callable] = {}

    def execute(self, method: str, args: Dict[str, object]):
        """
        Executes given method with given arguments

        :param method: method name to execute
        :param args: arguments to pass into method
        :return: method result
        """

        self._validate_args(method, args)
        return self.get_method(method)(**args)

    def _validate_args(self, method: str, args: Dict[str, object]):
        needed_args = self.exposed_method_args(method)
        missing_args = [arg.name for arg in needed_args if arg.name not in args]
        if len(missing_args) > 0:
            raise ExecutionError('{} method {} missing args {}'.format(self, method, ', '.join(missing_args)))

    def exposed_methods(self):
        """
        Lists signatures of methods exposed by interface

        :return: list of signatures
        """

        return list(self.exposed.keys())

    def get_method(self, method_name: str) -> callable:
        """Returns callable exposed method object with given name

        :param method_name: method name
        """
        try:
            return self.executors[method_name] if method_name in self.executors else getattr(self, method_name)
        except AttributeError:
            raise ExecutionError(f'Interface {self} does not have method "{method_name}"')

    def exposed_method_signature(self, method_name: str) -> Signature:
        """
        Gets signature of given method

        :param method_name: name of method to get signature for
        :return: signature
        """

        if method_name not in self.exposed:
            raise ExecutionError('Interface {} does not have method "{}"'.format(self, method_name))

        return self.exposed[method_name]

    def exposed_method_docs(self, method_name: str) -> str:
        """Gets docstring for given method

        :param method_name: name of the method
        :return: docstring
        """
        return getattr(self.get_method(method_name), '__doc__', None)

    def exposed_method_args(self, method_name: str) -> List[Field]:
        """
        Gets argument types of given method

        :param method_name: name of method to get argument types for
        :return: list of argument types
        """
        return self.exposed_method_signature(method_name).args

    def exposed_method_returns(self, method_name: str) -> Field:
        """
        Gets return type of given method

        :param method_name: name of method to get return type for
        :return: return type
        """

        return self.exposed_method_signature(method_name).output


class InterfaceMethodDescriptor(Comparable):
    """
    Descriptor of interface method: contains metadata only and couldn't execute method calls
    """

    # TODO support for types other then DatasetType, for example builtins
    def __init__(self, name: str, args: Dict[str, 'objects.DatasetType'], out_type: 'objects.DatasetType'):
        self.name = name
        self.args = args
        self.out_type = out_type

    @staticmethod
    def from_signature(name: str, signature: Signature):
        return InterfaceMethodDescriptor(name, {a.name: a.type for a in signature.args},
                                         signature.output.type)


# class _InterfaceMethodDescriptorSerializer(StaticSerializer):
#     real_type = InterfaceMethodDescriptor
#     @classmethod
#     def deserialize(cls, obj: dict) -> InterfaceMethodDescriptor:
#         return InterfaceMethodDescriptor(obj['name'],
#                                          {a['name']: deserialize(a['type'], DatasetType)for a in obj['args']},
#                                          deserialize(obj['out_type'], DatasetType))
#
#     @classmethod
#     def serialize(cls, instance: InterfaceMethodDescriptor) -> dict:


class InterfaceDescriptor(Comparable):
    """
    Descriptor of :class:`Interface`: contains metadata only and couldn't execute method calls
    """

    def __init__(self, methods: List[InterfaceMethodDescriptor], version: str):
        self.methods = methods
        self.version = version

    def to_dict(self):
        return serialize(self)

    @classmethod
    def from_dict(cls, d: dict):
        return deserialize(d, cls)

    @staticmethod
    def from_interface(interface: Interface):
        import algolink
        return InterfaceDescriptor(
            [InterfaceMethodDescriptor.from_signature(name, interface.exposed_method_signature(name))
             for name in interface.exposed_methods()], algolink.__version__)


_InterfaceLoaderBase = registering_type('loader')


class InterfaceLoader(_InterfaceLoaderBase):
    """
    Base class for loaders of :class:`Interface`
    """

    @abstractmethod
    def load(self) -> Interface:
        pass  # pragma: no cover

    @staticmethod
    def get(class_path) -> 'InterfaceLoader':
        return _InterfaceLoaderBase.get(class_path)()



import io
from functools import wraps

from imageio import imread, imsave


def bytes_image_input(f):
    """
    Decorator which marks that function consumes images stored as bytes objects.

    :param f: function to decorate
    :return: decorated function
    """
    @wraps(f)
    def inner(b):
        im = imread(b)
        return f(im)

    return inner


def bytes_image_output(f):
    """
    Decorator which marks that function returns images stored as bytes objects.

    :param f: function to decorate
    :return: decorated function
    """

    @wraps(f)
    def inner(*args, **kwargs):
        im = f(*args, **kwargs)
        buffer = io.BytesIO()
        imsave(buffer, im, format='PNG')
        return buffer.getvalue()

    return inner



from .. import config
from .interface import InterfaceLoader
from .server import Server
from ..utils.log import rlogger


def start_runtime(loader=None, server=None):
    """
    Starts AlgoLink runtime for given (optional) loader and (optional) server

    :param loader: loader of model to start AlgoLink runtime for,
        if not given class specified in :attr:`.config.Runtime.LOADER` is used
    :param server: server to use for AlgoLink runtime, default is a flask-based server,
        if not given class specified in :attr:`.config.Runtime.SERVER` is used
    :return: nothing
    """

    if not isinstance(server, Server):
        server = config.Runtime.SERVER
        server = Server.get(server)

    if not isinstance(loader, InterfaceLoader):
        loader = config.Runtime.LOADER
        loader = InterfaceLoader.get(loader)

    rlogger.info('Starting AlgoLink runtime with loader %s and server %s ...',
                 type(loader).__name__,
                 type(server).__name__)
    server.start(loader)



import sys
import time
from typing import Dict, Generator, Type

import docker.errors

from algolink.build.runner.base import RunnerBase
from algolink.utils.log import logger

from .base import DockerContainer, DockerEnv, DockerImage


class DockerRunnerException(Exception):
    pass


class DockerRunner(RunnerBase):
    """RunnerBase implementation for docker containers"""

    def instance_exists(self, instance: DockerContainer, env: DockerEnv, **kwargs) -> bool:
        with env.daemon.client() as client:
            try:
                client.containers.get(instance.name)
                return True
            except docker.errors.NotFound:
                return False

    def remove_instance(self, instance: DockerContainer, env: DockerEnv, **kwargs):
        with env.daemon.client() as client:
            try:
                c = client.containers.get(instance.name)
                c.remove(**kwargs)
            except docker.errors.NotFound:
                pass

    def instance_type(self) -> Type[DockerContainer]:
        return DockerContainer

    def create_instance(self, name: str, port_mapping: Dict[int, int] = None, **kwargs) -> DockerContainer:
        return DockerContainer(name, port_mapping, kwargs)

    def run(self, instance: DockerContainer, image: DockerImage, env: DockerEnv, rm=True, detach=True, **kwargs):
        if not (isinstance(instance, DockerContainer) and isinstance(image, DockerImage) and
                isinstance(env, DockerEnv)):
            raise TypeError('DockerRunner works with DockerContainer, DockerImage and DockerHost only')

        with env.daemon.client() as client:
            image.registry.login(client)

            try:
                # always detach from container and just stream logs if detach=False
                container = client.containers.run(image.uri,
                                                  name=instance.name,
                                                  auto_remove=rm,
                                                  ports=instance.port_mapping,
                                                  detach=True,
                                                  **instance.params,
                                                  **kwargs)
                instance.container_id = container.id
                if not detach:
                    try:
                        # infinite loop of logs while container running or if everything ok
                        for log in self._logs(container, stream=True):
                            logger.debug(log)

                        self._sleep()
                        if not self._is_service_running(client, instance.name):
                            raise DockerRunnerException("The container died unexpectedly.")

                    except KeyboardInterrupt:
                        logger.info('Interrupted. Stopping the container')
                        container.stop()

                else:
                    self._sleep(.5)
                    if not self._is_service_running(client, instance.name):
                        if not rm:
                            for log in self._logs(container, stdout=False, stderr=True):
                                raise DockerRunnerException("The container died unexpectedly.", log)
                        else:
                            # Can't get logs from removed container
                            raise DockerRunnerException("The container died unexpectedly. Try to run the container "
                                                        "with detach=False or rm=False args to get more info.")
            except docker.errors.ContainerError as e:
                if e.stderr:
                    print(e.stderr.decode(), file=sys.stderr)
                raise

    def logs(self, instance: DockerContainer, env: DockerEnv, **kwargs) -> Generator[str, None, None]:
        self._validate(instance, env)

        with env.daemon.client() as client:
            container = client.containers.get(instance.name)
            yield from self._logs(container, **kwargs)

    def _logs(self, container, stdout=True, stderr=True, stream=False,
              tail='all', since=None, follow=None, until=None, **kwargs) -> Generator[str, None, None]:

        log = container.logs(stdout=stdout, stderr=stderr, stream=stream,
                             tail=tail, since=since, follow=follow, until=until)
        if stream:
            for line in log:
                yield line.decode("utf-8")
        else:
            yield log.decode("utf-8")

    def _sleep(self, timeout: float = 10):
        time.sleep(timeout)

    def _is_service_running(self, client, name):
        from docker.errors import NotFound
        try:
            container = client.containers.get(name)
            return container.status == 'running'
        except NotFound:
            return False

    def is_running(self, instance: DockerContainer, env: DockerEnv, **kwargs) -> bool:
        self._validate(instance, env)

        with env.daemon.client() as client:
            return self._is_service_running(client, instance.name)

    def stop(self, instance: DockerContainer, env: DockerEnv, **kwargs):
        self._validate(instance, env)

        with env.daemon.client() as client:
            try:
                container = client.containers.get(instance.name)
                container.stop()
            except docker.errors.NotFound:
                pass

    @classmethod
    def _validate(cls, instance: DockerContainer, env: DockerEnv):
        if not (isinstance(instance, DockerContainer) and isinstance(env, DockerEnv)):
            raise TypeError('DockerRunner works with DockerContainer and DockerHost only')



import requests

from algolink.runtime.client.base import BaseClient
from algolink.runtime.interface.base import ExecutionError, InterfaceDescriptor


class HTTPClient(BaseClient):
    """
    Simple implementation of HTTP-based AlgiLink runtime client.

    Interface definition is acquired via HTTP GET call to `/interface.json`,
    method calls are performed via HTTP POST calls to `/<name>`.

    :param host: host of server to connect to, if no host given connects to host `localhost`
    :param port: port of server to connect to, if no port given connects to port 9000
    """

    def __init__(self, host=None, port=None):
        self.base_url = f'http://{host or "localhost"}:{port or 9000}'
        super().__init__()

    def _interface_factory(self) -> InterfaceDescriptor:
        resp = requests.get(f'{self.base_url}/interface.json')
        resp.raise_for_status()
        return InterfaceDescriptor.from_dict(resp.json())

    def _call_method(self, name, args):
        ret = requests.post(f'{self.base_url}/{name}', json=args)
        if ret.status_code == 200:
            return ret.json()['data']
        elif ret.status_code == 400:
            raise ExecutionError(ret.json()['error'])  # TODO tests
        else:
            ret.raise_for_status()



import contextlib
import os
import typing
from io import BytesIO

import torch
from pyjackson.decorators import make_string

from algolink.core.analyzer import TypeHookMixin
from algolink.core.analyzer.model import BindingModelHook
from algolink.core.objects.artifacts import ArtifactCollection, Blobs, InMemoryBlob
from algolink.core.objects.wrapper import ModelIO, ModelWrapper


class TorchModelIO(ModelIO):
    """
    :class:`algolink.core.objects.ModelIO` for PyTorch models
    """
    model_file_name = 'model.pth'
    model_jit_file_name = 'model.jit.pth'

    @contextlib.contextmanager
    def dump(self, model) -> ArtifactCollection:
        """
        Dumps `torch.nn.Module` instance to :class:`.InMemoryBlob` and creates :class:`.ArtifactCollection` from it

        :return: context manager with :class:`~algolink.core.objects.ArtifactCollection`
        """
        is_jit = isinstance(model, torch.jit.ScriptModule)
        save = torch.jit.save if is_jit else torch.save
        model_name = self.model_jit_file_name if is_jit else self.model_file_name

        buffer = BytesIO()
        save(model, buffer)
        yield Blobs({model_name: InMemoryBlob(buffer.getvalue())})

    def load(self, path):
        """
        Loads `torch.nn.Module` instance from path

        :param path: path to load from
        """
        model_path = os.path.join(path, self.model_jit_file_name)
        load = torch.jit.load
        if not os.path.exists(model_path):
            model_path = os.path.join(path, self.model_file_name)
            load = torch.load

        with open(model_path, 'rb') as f:
            return load(f)


class TorchModelWrapper(ModelWrapper):
    """
    :class:`algolink.core.objects.ModelWrapper` for PyTorch models. `.model` attribute is a `torch.nn.Module` instance
    """
    def __init__(self):
        super().__init__(TorchModelIO())

    def _exposed_methods_mapping(self) -> typing.Dict[str, str]:
        return {
            'predict': '_predict'
        }

    @ModelWrapper.with_model
    def _predict(self, data):
        if isinstance(data, torch.Tensor):
            return self.model(data)
        return self.model(*data)


@make_string(include_name=True)
class TorchModelHook(BindingModelHook, TypeHookMixin):
    """
    Hook for PyTorch models
    """
    valid_types = [torch.nn.Module]

    def _wrapper_factory(self) -> ModelWrapper:
        """
        Creates :class:`TorchModelWrapper` for PyTorch model object

        :return: :class:`TorchModelWrapper` instance
        """
        return TorchModelWrapper()



from abc import abstractmethod
from functools import wraps
from typing import List, Optional, Sequence, TypeVar, Union

from pyjackson.decorators import type_field

from algolink.client.expose import ExposedMethod
from algolink.core import errors
from algolink.core.objects import core

Project = 'core.Project'
Task = 'core.Task'
Model = 'core.Model'
Image = 'core.Image'
Pipeline = 'core.Pipeline'
RuntimeEnvironment = 'core.RuntimeEnvironment'

## ML tracking
Experiment = 'core.Experiment'
ModelMetric = 'core.ModelMetric'
ModelParam ='core.ModelParam'
BestResult = 'core.BestResult'

T = TypeVar('T')
NameOrIdOrObject = Union[int, str, T]
ProjectVar = NameOrIdOrObject[Project]
TaskVar = NameOrIdOrObject[Task]
ModelVar = NameOrIdOrObject[Model]
PipelineVar = NameOrIdOrObject[Pipeline]
EnvironmentVar = NameOrIdOrObject[RuntimeEnvironment]

ExperimentVar = NameOrIdOrObject[Experiment]
ModelMetricVar = NameOrIdOrObject[ModelMetric]
ModelParamVar = NameOrIdOrObject[ModelParam]
BestResultVar = NameOrIdOrObject[BestResult]

class ExposedMetadataMethod(ExposedMethod):
    def __init__(self, bind=True, name: str = None):
        super().__init__(name)
        self.bind = bind
        self.method = None

    def generate_code(self):
        fields = self.get_signature().args
        declaration = self.get_declaration().replace('core.', '')
        result = f'''self.meta_repo.{self.original_name}({', '.join(f.name for f in fields)})'''
        if self.bind:
            result = f'''self._bind({result})'''
        return f'''    {declaration}
        """{self.method.__doc__}"""
        return {result}'''


def bind_to_self(method):
    """
    Decorator for methods which binds method result to metadata repository contained in `self` reference.

    :param method: method to decorate
    :return: decorated method
    """

    @wraps(method)
    def inner(self, *args, **kwargs):
        res = method(self, *args, **kwargs)
        if isinstance(res, core.AlgoLinkObject):
            res.bind_meta_repo(self)
        elif isinstance(res, Sequence):
            for o in res:
                o.bind_meta_repo(self)
        return res

    return inner


@type_field('type')
class MetadataRepository:
    """
    Abstract base class for persistent repositories of metadata (:class:`.core.Project`, :class:`.core.Task`, etc)
    """

    type = None

    @abstractmethod
    @ExposedMetadataMethod()
    def get_projects(self) -> List['core.Project']:
        """
        Gets all projects in the repository

        :return: all projects in the repository
        """

    @abstractmethod
    @ExposedMetadataMethod(name='get_project')
    def get_project_by_name(self, name: str) -> Optional['core.Project']:
        """
        Finds project in the repository by name

        :param name: name of the project to return
        :return: found project if exists or `None`
        """
    
    @abstractmethod
    def get_project_by_id(self, id: int) -> Optional['core.Project']:
        """
        Finds project in the repository by identifier

        :param id: project id
        :return: found project if exists or `None`
        """

    @abstractmethod
    def create_project(self, project: Project) -> Project:
        """
        Creates the project and all its tasks.

        :param project: project to create
        :return: created project
        :exception: :exc:`.errors.ExistingProjectError` if given project has the same name as existing one.
        """
    @abstractmethod
    def update_project(self, project: Project) -> Project:
        """
        Updates the project and all its tasks.

        :param project: project to update
        :return: updated project
        :exception: :exc:`.errors.NonExistingProjectError` if given project doesn't exist in the repository
        """

    @abstractmethod
    def delete_project(self, project: Project):
        """
        Deletes the project and all tasks.

        :param project: project to delete
        :return: nothing
        :exception: :exc:`.errors.NonExistingProjectError` if given project doesn't exist in the repository
        """

    def save_project(self, project: Project) -> Project:
        """
        Saves project into the repository

        :param project: project to save
        :return: saved project
        :exception: :exc:`.errors.ExistingProjectError` if given project has the same name as existing one.
        """
        existing_project = self.get_project_by_name(project.name)
        if project.id is None and existing_project is None:
            return self.create_project(project)
        elif existing_project is not None:
            if project.id is None or existing_project.id != project.id:
                raise errors.ExistingProjectError(existing_project)
        return self.update_project(project)

    @ExposedMetadataMethod()
    def get_or_create_project(self, name: str) -> Project:
        """
        Creates a project if not exists or gets existing project otherwise.

        :param name: project name
        :return: project
        """
        project = self.get_project_by_name(name)
        if project is None:
            project = core.Project(name)
            project = self.create_project(project)
        return project     
    
    @abstractmethod
    @ExposedMetadataMethod()
    def get_tasks(self, project: ProjectVar) -> List['core.Task']:
        """
        Gets a list of tasks for given project

        :param project: project to search for tasks in
        :return: project tasks
        """

    @abstractmethod
    @ExposedMetadataMethod(name='get_task')
    def get_task_by_name(self, project: ProjectVar, task_name: str) -> Optional['core.Task']:
        """
        Finds task with given name in given project

        :param project: project to search for task in
        :param task_name: expected name of task
        :return: task if exists or `None`
        """

    @abstractmethod
    def get_task_by_id(self, id: int) -> Optional['core.Task']:
        """
        Finds task with given id

        :param id: id of task to search for
        :return: task if exists or `None`
        """

    @ExposedMetadataMethod()
    def get_or_create_task(self, project: str, task_name: str) -> Task:
        """
        Creates a task if not exists or gets existing task otherwise.

        :param project: project to search/create task in
        :param task_name: expected name of task
        :return: created/found task
        """

        project = self.get_or_create_project(project)
        task = self.get_task_by_name(project, task_name)
        if task is None:
            task = core.Task(task_name, project_id=project.id)
            task = self.create_task(task)
        return task

    @abstractmethod
    def create_task(self, task: Task) -> Task:
        """
        Creates task in a repository

        :param task: task to create
        :return: created task
        :exception: :class:`.errors.ExistingTaskError` if given task has the same name and project as existing one
        """

    @abstractmethod
    def update_task(self, task: Task) -> Task:
        """
        Updates task in a repository.

        :param task: task to update
        :return: updated task
        :exception: :exc:`.errors.NonExistingTaskError` if given tasks doesn't exist in the repository
        """

    @abstractmethod
    def delete_task(self, task: Task):
        """
        Deletes the task and all its models.

        :param task: task to delete
        :return: nothing
        :exception: :exc:`.errors.NonExistingTaskError` if given tasks doesn't exist in the repository
        """

    def save_task(self, task: Task) -> Task:
        """
        Saves task into repository

        :param task: task
        :return: saved task
        :exception: :class:`.errors.ExistingTaskError` if given task has the same name and project as existing one
        """

        if task.project_id is None:
            raise ValueError("A project is not assigned to the task {}".format(task))

        existing_task = self.get_task_by_name(task.project_id, task.name)
        if task.id is None and existing_task is None:
            return self.create_task(task)
        elif existing_task is not None:
            if task.id is None or existing_task.id != task.id:
                raise errors.ExistingTaskError(existing_task)
        return self.update_task(task)

    @abstractmethod
    @ExposedMetadataMethod()
    def get_models(self, task: TaskVar, project: ProjectVar = None) -> List['core.Model']:
        """
        Gets a list of models in given project and task

        :param task: task to search for models in
        :param project: project to search for models in
        :return: found models
        """

    @abstractmethod
    def get_model_by_name(self, model_name: str, task: TaskVar, project: ProjectVar = None) -> Optional['core.Model']:
        """
        Finds model by name in given task and project.

        :param model_name: expected model name
        :param task: task to search for model in
        :param project: project to search for model in
        :return: found model if exists or `None`
        """

    @abstractmethod
    def get_model_by_id(self, id: int) -> Optional['core.Model']:
        """
        Finds model by identifier.

        :param id: expected model id
        :return: found model if exists or `None`
        """

    @abstractmethod
    def create_model(self, model: Model) -> Model:
        """
        Creates model in the repository

        :param model: model to create
        :return: created model
        :exception: :exc:`.errors.ExistingModelError` if given model has the same name and task as existing one
        """

    @abstractmethod
    def update_model(self, model: Model) -> Model:
        """
        Updates model in the repository

        :param model: model to update
        :return: updated model
        :exception: :exc:`.errors.NonExistingModelError` if given model doesn't exist in the repository
        """

    @abstractmethod
    def delete_model(self, model: Model):
        """
        Deletes model from the repository

        :param model: model to delete
        :return: nothing
        :exception: :exc:`.errors.NonExistingModelError` if given model doesn't exist in the repository
        """

    def save_model(self, model: Model) -> Model:
        """
        Saves model in the repository

        :param model: model to save
        :return: saved model
        :exception: :exc:`.errors.ExistingModelError` if given model has the same name and task as existing one
        """

        if model.task_id is None:
            raise ValueError("A task is not assigned to the model {}".format(model))

        existing_model = self.get_model_by_name(model.name, model.task_id)

        if model.id is None and existing_model is None:
            return self.create_model(model)
        elif existing_model is not None:
            if model.id is None or existing_model.id != model.id:
                raise errors.ExistingModelError(model)
        return self.update_model(model)

    # ___________________

    @abstractmethod
    @ExposedMetadataMethod()
    def get_pipelines(self, task: TaskVar, project: ProjectVar = None) -> List['core.Pipeline']:
        """
        Gets a list of pipelines in given project and task

        :param task: task to search for models in
        :param project: project to search for models in
        :return: found pipelines
        """

    @abstractmethod
    @ExposedMetadataMethod(name='get_pipeline')
    def get_pipeline_by_name(self, pipeline_name: str, task: TaskVar,
                             project: ProjectVar = None) -> Optional['core.Pipeline']:
        """
        Finds model by name in given task and project.

        :param pipeline_name: expected pipeline name
        :param task: task to search for pipeline in
        :param project: project to search for pipeline in
        :return: found pipeline if exists or `None`
        """

    @abstractmethod
    def get_pipeline_by_id(self, id: int) -> Optional['core.Pipeline']:
        """
        Finds model by identifier.

        :param id: expected model id
        :return: found model if exists or `None`
        """

    @abstractmethod
    def create_pipeline(self, pipeline: Pipeline) -> Pipeline:
        """
        Creates model in the repository

        :param pipeline: pipeline to create
        :return: created pipeline
        :exception: :exc:`.errors.ExistingPipelineError` if given model has the same name and task as existing one
        """

    @abstractmethod
    def update_pipeline(self, pipeline: Pipeline) -> Pipeline:
        """
        Updates model in the repository

        :param pipeline: pipeline to update
        :return: updated model
        :exception: :exc:`.errors.NonExistingPipelineError` if given pipeline doesn't exist in the repository
        """

    @abstractmethod
    def delete_pipeline(self, pipeline: Pipeline):
        """
        Deletes model from the repository

        :param pipeline: pipeline to delete
        :return: nothing
        :exception: :exc:`.errors.NonExistingPipelineError` if given pipeline doesn't exist in the repository
        """

    def save_pipeline(self, pipeline: Pipeline) -> Pipeline:
        """
        Saves pipeline in the repository

        :param pipeline: model to save
        :return: saved pipeline
        :exception: :exc:`.errors.ExistingPipelineError` if given pipeline has the same name and task as existing one
        """

        if pipeline.task_id is None:
            raise ValueError("A task is not assigned to the pipeline {}".format(pipeline))

        existing_pipeline = self.get_pipeline_by_name(pipeline.name, pipeline.task_id)

        if pipeline.id is None and existing_pipeline is None:
            return self.create_pipeline(pipeline)
        elif existing_pipeline is not None:
            if pipeline.id is None or existing_pipeline.id != pipeline.id:
                raise errors.ExistingPipelineError(pipeline)
        return self.update_pipeline(pipeline)

    # _______________
    @abstractmethod
    @ExposedMetadataMethod()
    def get_images(self, task: TaskVar, project: ProjectVar = None) -> List['core.Image']:
        """
        Gets a list of images in given model, task and project

        :param task: task to search for images in
        :param project: project to search for images in
        :return: found images
        """

    @abstractmethod
    @ExposedMetadataMethod(name='get_image')
    def get_image_by_name(self, image_name: str, task: TaskVar, project: ProjectVar = None) -> Optional['core.Image']:
        """
        Finds image by name in given model, task and project.

        :param image_name: expected image name
        :param task: task to search for image in
        :param project: project to search for image in
        :return: found image if exists or `None`
        """

    @abstractmethod
    def get_image_by_id(self, id: int) -> Optional['core.Image']:
        """
        Finds image by identifier.

        :param id: expected image id
        :return: found image if exists or `None`
        """

    @abstractmethod
    def create_image(self, image: Image) -> Image:
        """
        Creates image in the repository

        :param image: image to create
        :return: created image
        :exception: :exc:`.errors.ExistingImageError` if given image has the same name and model as existing one
        """

    @abstractmethod
    def update_image(self, image: Image) -> Image:
        """
        Updates image in the repository

        :param image: image to update
        :return: updated image
        :exception: :exc:`.errors.NonExistingImageError` if given image doesn't exist in the repository
        """

    @abstractmethod
    def delete_image(self, image: Image):
        """
        Deletes image from the repository

        :param image: image to delete
        :return: nothing
        :exception: :exc:`.errors.NonExistingImageError` if given image doesn't exist in the repository
        """

    def save_image(self, image: Image) -> Image:
        """
        Saves image in the repository

        :param image: image to save
        :return: saved image
        :exception: :exc:`.errors.ExistingImageError` if given image has the same name and model as existing one
        """
        self._validate_image(image)

        existing_image = self.get_image_by_name(image.name, image.task_id)

        if image.id is None and existing_image is None:
            return self.create_image(image)
        elif existing_image is not None:
            if image.id is None or existing_image.id != image.id:
                raise errors.ExistingImageError(image)
        return self.update_image(image)

    @abstractmethod
    @ExposedMetadataMethod()
    def get_environments(self) -> List[RuntimeEnvironment]:
        """
        Gets a list of runtime environments

        :return: found runtime environments
        """

    @abstractmethod
    @ExposedMetadataMethod(name='get_environment')
    def get_environment_by_name(self, name: str) -> Optional[RuntimeEnvironment]:
        """
        Finds runtime environment by name.

        :param name: expected runtime environment name
        :return: found runtime environment if exists or `None`
        """

    @abstractmethod
    def get_environment_by_id(self, id: int) -> Optional[RuntimeEnvironment]:
        """
        Finds runtime environment by identifier.

        :param id: expected runtime environment id
        :return: found runtime environment if exists or `None`
        """

    @abstractmethod
    @ExposedMetadataMethod(name='push_environment')
    def create_environment(self, environment: 'core.RuntimeEnvironment') -> RuntimeEnvironment:
        """
        Creates runtime environment in the repository

        :param environment: runtime environment to create
        :return: created runtime environment
        :exception: :exc:`.errors.ExistingEnvironmentError` if given runtime environment has the same name as existing
        """

    @abstractmethod
    def update_environment(self, environment: 'core.RuntimeEnvironment') -> RuntimeEnvironment:
        """
        Updates runtime environment in the repository

        :param environment: runtime environment to update
        :return: updated runtime environment
        :exception: :exc:`.errors.NonExistingEnvironmentError` if given runtime environment doesn't exist in the
        repository
        """

    @abstractmethod
    def delete_environment(self, environment: 'core.RuntimeEnvironment'):
        """
        Deletes runtime environment from the repository

        :param environment: runtime environment to delete
        :return: nothing
        :exception: :exc:`.errors.NonExistingEnvironmentError` if given runtime environment doesn't exist in the
        repository
        """

    def save_environment(self, environment: 'core.RuntimeEnvironment') -> RuntimeEnvironment:
        """
        Saves runtime environment in the repository

        :param environment: runtime environment to save
        :return: saved runtime environment
        :exception: :exc:`.errors.ExistingEnvironmentError` if given runtime environment has the same name as existing
        """
        self._validate_environment(environment)

        existing_environment = self.get_environment_by_name(environment.name)

        if environment.id is None and existing_environment is None:
            return self.create_environment(environment)
        elif existing_environment is not None:
            if environment.id is None or existing_environment.id != environment.id:
                raise errors.ExistingEnvironmentError(environment)
        return self.update_environment(environment)

    @abstractmethod
    @ExposedMetadataMethod()
    def get_instances(self, image: Union[int, 'core.Image'] = None,
                      environment: Union[int, 'core.RuntimeEnvironment'] = None) -> List['core.RuntimeInstance']:
        """
        Gets a list of instances in given image or environment

        :param image: image (or id) to search for instances in
        :param environment: environment (or id) to search for instances in
        :return: found instances
        """

    @abstractmethod
    @ExposedMetadataMethod(name='get_instance')
    def get_instance_by_name(self, instance_name: str, image: Union[int, 'core.Image'],
                             environment: Union[int, 'core.RuntimeEnvironment']) -> Optional['core.RuntimeInstance']:
        """
        Finds instance by name in given image and environment.

        :param instance_name: expected instance name
        :param image: image (or id) to search for instance in
        :param environment: environment (or id) to search for instance in
        :return: found instance if exists or `None`
        """

    @abstractmethod
    def get_instance_by_id(self, id: int) -> Optional['core.RuntimeInstance']:
        """
        Finds instance by identifier.

        :param id: expected instance id
        :return: found instance if exists or `None`
        """

    @abstractmethod
    def create_instance(self, instance: 'core.RuntimeInstance') -> 'core.RuntimeInstance':
        """
        Creates instance in the repository

        :param instance: instance to create
        :return: created instance
        :exception: :exc:`.errors.ExistingInstanceError` if given instance has the same name, image and environment
            as existing one
        """

    @abstractmethod
    def update_instance(self, instance: 'core.RuntimeInstance') -> 'core.RuntimeInstance':
        """
        Updates instance in the repository

        :param instance: instance to update
        :return: updated instance
        :exception: :exc:`.errors.NonExistingInstanceError` if given instance doesn't exist in the repository
        """

    @abstractmethod
    def delete_instance(self, instance: 'core.RuntimeInstance'):
        """
        Deletes instance from the repository

        :param instance: instance to delete
        :return: nothing
        :exception: :exc:`.errors.NonExistingInstanceError` if given instance doesn't exist in the repository
        """

    def save_instance(self, instance: 'core.RuntimeInstance') -> 'core.RuntimeInstance':
        """
        Saves instance in the repository

        :param instance: instance to save
        :return: saved instance
        :exception: :exc:`.errors.ExistingInstanceError` if given image has the same name, image and environment
            as existing one
        """
        self._validate_instance(instance)

        existing_instance = self.get_instance_by_name(instance.name, instance.image_id, instance.environment_id)

        if instance.id is None and existing_instance is None:
            return self.create_instance(instance)
        elif existing_instance is not None:
            if instance.id is None or existing_instance.id != instance.id:
                raise errors.ExistingInstanceError(instance)
        return self.update_instance(instance)

    def _resolve_project(self, project: ProjectVar) -> Optional['core.Project']:
        if isinstance(project, core.Project):
            project = project.id if project.id is not None else project.name
        if isinstance(project, int):
            return self.get_project_by_id(project)
        return self.get_project_by_name(project)

    def _resolve_task(self, task: TaskVar, project: ProjectVar = None) -> Optional['core.Task']:
        if isinstance(task, core.Task):
            task = task.id if task.id is not None else task.name
        if isinstance(task, int):
            return self.get_task_by_id(task)
        if project is None:
            raise ValueError('Cannot resolve task without project')
        return self.get_task_by_name(project, task)

    def _resolve_model(self, model: ModelVar, task: TaskVar = None, project: ProjectVar = None):
        if isinstance(model, core.Model):
            model = model.id if model.id is not None else model.name
        if isinstance(model, int):
            return self.get_model_by_id(model)
        if task is None:
            raise ValueError('Cannot resolve model without task')
        return self.get_model_by_name(model, task, project)

    def _resolve_environment(self, environment: EnvironmentVar) -> Optional[RuntimeEnvironment]:
        if isinstance(environment, core.RuntimeEnvironment):
            environment = environment.id if environment.id is not None else environment.name
        if isinstance(environment, int):
            return self.get_environment_by_id(environment)
        return self.get_environment_by_name(environment)

    def _validate_project(self, project: Project):
        pass

    def _validate_task(self, task: Task):
        if task.project_id is None:
            raise errors.TaskNotInProjectError(task)

    def _validate_experiment(self, experiment: Experiment):
        if experiment.task_id is None:
            raise errors.ExperimentNotInTaskError(experiment)

    def _validate_model(self, model: Model):
        if model.task_id is None:
            raise errors.ModelNotInTaskError(model)

    def _validate_pipeline(self, pipeline: Pipeline):
        if pipeline.task_id is None:
            raise errors.PipelineNotInTaskError(pipeline)

    def _validate_image(self, image: Image):
        if image.task_id is None:
            raise errors.ImageNotInTaskError(image)

    def _validate_environment(self, environment: 'core.RuntimeEnvironment'):
        pass

    def _validate_instance(self, instance: 'core.RuntimeInstance'):
        if instance.image_id is None:
            raise errors.InstanceNotInImageError(instance)
        if instance.environment_id is None:
            raise errors.InstanceNotInEnvironmentError(instance)

    @abstractmethod
    @ExposedMetadataMethod()
    def get_experiments(self, task: TaskVar) -> List['core.Experiment']:
        """
        Gets a list of experiments for given task

        :param task: task to search for experiments in
        :return: task experiments
        """
    @abstractmethod
    @ExposedMetadataMethod(name='get_experiment')
    def get_experiment_by_name(self, task: TaskVar, experiment_name: str) -> Optional['core.Experiment']:
        """
        Finds experiment with given name in given task

        :param task: task to search for experiment in
        :param experiment_name: expected name of experiment
        :return: experiment if exists or `None`
        """
    @abstractmethod
    @ExposedMetadataMethod(name='get_experiment_sequence')
    def get_experiment_sequence(self,task_id: str)-> Optional['core.Experiment']:
        """
        Finds experiment count with given name in given task

        :param task: task to search for experiment count in
        :param task_id: expected id of task
        :return: experiment count if exists or `None`
        """
    @abstractmethod
    def get_experiment_by_id(self, id: int) -> Optional['core.Experiment']:
        """
        Finds experiment with given id

        :param id: id of experiment to search for
        :return: experiment if exists or `None`
        """

    @ExposedMetadataMethod()
    def get_or_create_experiment(self,project: str, task: str, experiment_name: str) -> Experiment:
        """
        Creates a experiment if not exists or gets existing experiment otherwise.

        :param task: task to search/create experiment in
        :param experiment_name: expected name of experiment
        :return: created/found experiment
        """

        task = self.get_or_create_task(project,task)
        experiment = self.get_experiment_by_name(task, experiment_name)
        #print(task,experiment,task.id,"experiment")
        if experiment is None:
            experiment_sequence = self.get_experiment_sequence(task.id)
            experiment = core.Experiment(experiment_name, experiment_sequence= experiment_sequence+1, task_id=task.id)
            experiment = self.create_experiment(experiment)
            
        return experiment

    @abstractmethod
    def create_experiment(self, experiment: Experiment) -> Experiment:
        """
        Creates experiment in a repository

        :param experiment: experiment to create
        :return: created experiment
        :exception: :class:`.errors.ExistingExperimentError` if given experiment has the same name and task as existing one
        """

    @abstractmethod
    def update_experiment(self, experiment: Experiment) -> Experiment:
        """
        Updates experiment in a repository.

        :param experiment: experiment to update
        :return: updated experiment
        :exception: :exc:`.errors.NonExistingExperimentError` if given experiments doesn't exist in the repository
        """

    @abstractmethod
    def delete_experiment(self, experiment: Experiment):
        """
        Deletes the experiment and all its models.

        :param experiment: experiment to delete
        :return: nothing
        :exception: :exc:`.errors.NonExistingExperimentError` if given experiments doesn't exist in the repository
        """

    def save_experiment(self, experiment: Experiment) -> Experiment:
        """
        Saves experiment into repository

        :param experiment: experiment
        :return: saved experiment
        :exception: :class:`.errors.ExistingExperimentError` if given experiment has the same name and experiment as existing one
        """

        if experiment.task_id is None:
            raise ValueError("A task is not assigned to the experiment {}".format(experiment))

        existing_experiment = self.get_experiment_by_name(experiment.task_id, experiment.name)
        if experiment.id is None and existing_experiment is None:
            return self.create_experiment(experiment)
        elif existing_experiment is not None:
            if experiment.id is None or existing_experiment.id != experiment.id:
                raise errors.ExistingExperimentError(existing_experiment)
        return self.update_experiment(experiment)
    
    
    
    def save_metric(self, metric: ModelMetric) -> ModelMetric:
        """
        Saves model metric into repository

        :param metric: metric
        :return: saved metric
        :exception: :class:`.errors.ExistingModelMetricError` if given metric has the same name and metric as existing one
        """

        if metric.experiment_id is None:
            raise ValueError("A experiment is not assigned to the metric {}".format(metric))
        existing_epoch = self.get_metric_by_name_epoch(metric.experiment_id, metric.name, metric.epoch)
        #print(existing_epoch,"existing_epoch,",metric.experiment_id, metric.name, metric.epoch)
        if existing_epoch is not None:
            return
            #raise errors.ExistingModelMetricError("epoch {}".format(existing_epoch))
        return self.create_metric(metric)
        #existing_metric = self.get_metric_by_name(metric.experiment_id, metric.name)
        #if metric.id is None and existing_metric is None:
        #    return self.create_metric(metric)
        #elif existing_metric is not None:
        #    if metric.id is None or existing_metric.id != metric.id:
        #        raise errors.ExistingModelMetricError(existing_metric)
        #return self.update_metric(metric)

    @abstractmethod
    @ExposedMetadataMethod(name='get_metric')
    def get_metric_by_name(self, experiment: ExperimentVar, metric_name: str) -> Optional['core.ModelMetric']:
        """
        Finds metric with given name in given experiment

        :param experiment: experiment to search for metric in
        :param metric_name: expected name of metric
        :return: metric if exists or `None`
        """
        
    @abstractmethod
    def create_metric(self, metric: ModelMetric) -> ModelMetric:
        """
        Creates metric in a repository

        :param metric: metric to create
        :return: created metric
        :exception: :class:`.errors.ExistingMetricError` if given metric has the same name and experiment as existing one
        """

    @abstractmethod
    def update_metric(self, metric: ModelMetric) -> ModelMetric:
        """
        Updates metric in a repository.

        :param metric: metric to update
        :return: updated metric
        :exception: :exc:`.errors.NonExistingMetricError` if given metrics doesn't exist in the repository
        """
        
    def save_param(self, param: ModelParam) -> ModelParam:
        """
        Saves model param into repository

        :param param: param
        :return: saved param
        :exception: :class:`.errors.ExistingModelParamError` if given metric has the same name and metric as existing one
        """

        if param.experiment_id is None:
            raise ValueError("A experiment is not assigned to the metric {}".format(param))

        return self.create_param(param)  

    def save_best_result(self,best_result: BestResult) -> BestResult:
        """
        Saves best result into repository

        :param param: best_result
        :return: saved best result
        :exception: :class:`.errors.ExistingBestResultError` if given best result has the same name and best result as existing one
        """

        if best_result.experiment_id is None:
            raise ValueError("A experiment is not assigned to the best result {}".format(best_result))

        return self.create_best_result(best_result)          

    @abstractmethod
    def create_param(self, param: ModelParam) -> ModelParam:
        """
        Creates param in a repository

        :param param: param to create
        :return: created param
        :exception: :class:`.errors.ExistingModelMetricError` if given param has the same name and experiment as existing one
        """

    @abstractmethod
    def create_best_result(self, best_result: BestResult) -> BestResult:
        """
        Creates best result in a repository

        :param param: best_result to create
        :return: created best_result
        :exception: :class:`.errors.ExistingBestResultError` if given best result has the same name and best result as existing one
        """


from typing import Union

from algolink.core.objects import (Image, Model, Pipeline, Project, RuntimeEnvironment, RuntimeInstance, Task,
                                   Experiment,ModelMetric)

ProjectIntStr = Union[Project, int, str]
TaskIntStr = Union[Task, int, str]
ModelIntStr = Union[Model, int, str]
PipelineIntStr = Union[Pipeline, int, str]
RuntimeEnvironmentIntStr = Union[RuntimeEnvironment, int, str]
RuntimeInstanceIntStr = Union[RuntimeInstance, int, str]
ImageIntStr = Union[Image, int, str]

ExperimentIntStr = Union[Experiment, int, str]
ModelMetricIntStr = Union[ModelMetric, int, str]

class AlgoLinkError(Exception):
    """
    General AlgoLink error
    """


class MetadataError(AlgoLinkError):
    """
    General AlgoLink Metadata Error
    """


class ExistingProjectError(MetadataError):
    def __init__(self, project: ProjectIntStr):
        project = project.name if isinstance(project, Project) else project
        super(ExistingProjectError, self).__init__('Project with name "{}" already exists'.format(project))


class NonExistingProjectError(MetadataError):
    def __init__(self, project: ProjectIntStr):
        project = project.name if isinstance(project, Project) else project
        if isinstance(project, int):
            super(NonExistingProjectError, self).__init__('Project with id "{}" does not exist'.format(project))
        else:
            super(NonExistingProjectError, self).__init__('Project with name "{}" does not exist'.format(project))

class NonExistingExperimentError(MetadataError):
    def __init__(self, experiment: ExperimentIntStr):
        experiment = experiment.name if isinstance(experiment, Experiment) else experiment
        if isinstance(experiment, int):
            super(NonExistingExperimentError, self).__init__('Experiment with id "{}" does not exist'.format(experiment))
        else:
            super(NonExistingexperimentError, self).__init__('Experiment with name "{}" does not exist'.format(experiment))

class ExistingTaskError(MetadataError):
    def __init__(self, task: TaskIntStr):
        task = task.name if isinstance(task, Task) else task
        super(ExistingTaskError, self).__init__('Task with name "{}" already exists'.format(task))
class ExistingExperimentError(MetadataError):
    def __init__(self, experiment: ExperimentIntStr):
        experiment = experiment.name if isinstance(experiment, Experiment) else experiment
        super(ExistingExperimentError, self).__init__('Experiment with name "{}" already exists'.format(experiment))
        
class ExistingModelMetricError(MetadataError):
    def __init__(self, modelmetric: ModelMetricIntStr):
        modelmetric = modelmetric.name if isinstance(modelmetric, ModelMetric) else modelmetric
        super(ExistingModelMetricError, self).__init__('ModelMetric with name "{}" already exists'.format(modelmetric))

class NonExistingTaskError(MetadataError):
    def __init__(self, task: TaskIntStr):
        task = task.name if isinstance(task, Task) else task
        if isinstance(task, int):
            super(NonExistingTaskError, self).__init__('Task with id "{}" does not exist'.format(task))
        else:
            super(NonExistingTaskError, self).__init__('Task with name "{}" does not exist'.format(task))


class TaskWithoutIdError(MetadataError):
    def __init__(self, task: TaskIntStr):
        task = task.name if isinstance(task, Task) else task
        super(TaskWithoutIdError, self).__init__('Task "{}" id is None'.format(task))


class ExistingModelError(MetadataError):
    def __init__(self, model: ModelIntStr):
        model = model.name if isinstance(model, Model) else model
        super(ExistingModelError, self).__init__('Model with name "{}" already exists'.format(model))


class NonExistingModelError(MetadataError):
    def __init__(self, model: ModelIntStr):
        model = model.name if isinstance(model, Model) else model
        if isinstance(model, int):
            super(NonExistingModelError, self).__init__('Model with id "{}" does not exist'.format(model))
        else:
            super(NonExistingModelError, self).__init__('Model with name "{}" does not exist'.format(model))


class ExistingPipelineError(MetadataError):
    def __init__(self, pipeline: PipelineIntStr):
        pipeline = pipeline.name if isinstance(pipeline, Pipeline) else pipeline
        super(ExistingPipelineError, self).__init__('Pipeline with name "{}" already exists'.format(pipeline))


class NonExistingPipelineError(MetadataError):
    def __init__(self, pipeline: PipelineIntStr):
        pipeline = pipeline.name if isinstance(pipeline, Model) else pipeline
        if isinstance(pipeline, int):
            super(NonExistingPipelineError, self).__init__('Pipeline with id "{}" does not exist'.format(pipeline))
        else:
            super(NonExistingPipelineError, self).__init__('Pipeline with name "{}" does not exist'.format(pipeline))


class ExistingImageError(MetadataError):
    def __init__(self, image: ImageIntStr):
        image = image.name if isinstance(image, Image) else image
        super(ExistingImageError, self).__init__('Image with name "{}" already exists'.format(image))


class NonExistingImageError(MetadataError):
    def __init__(self, image: ImageIntStr):
        image = image.name if isinstance(image, Image) else image
        if isinstance(image, int):
            super(NonExistingImageError, self).__init__('Image with id "{}" does not exist'.format(image))
        else:
            super(NonExistingImageError, self).__init__('Image with name "{}" does not exist'.format(image))


class ExistingEnvironmentError(MetadataError):
    def __init__(self, environment: RuntimeEnvironmentIntStr):
        environment = environment.name if isinstance(environment, RuntimeEnvironment) else environment
        super(ExistingEnvironmentError, self).__init__('Environment with name "{}" already exists'.format(environment))


class NonExistingEnvironmentError(MetadataError):
    def __init__(self, environment: RuntimeEnvironmentIntStr):
        environment = environment.name if isinstance(environment, RuntimeEnvironment) else environment
        if isinstance(environment, int):
            super(NonExistingEnvironmentError, self).__init__(
                'Environment with id "{}" does not exist'.format(environment))
        else:
            super(NonExistingEnvironmentError, self).__init__(
                'Environment with name "{}" does not exist'.format(environment))


class ExistingInstanceError(MetadataError):
    def __init__(self, instance: RuntimeInstanceIntStr):
        instance = instance.name if isinstance(instance, RuntimeInstance) else instance
        super(ExistingInstanceError, self).__init__('Instance with name "{}" already exists'.format(instance))


class NonExistingInstanceError(MetadataError):
    def __init__(self, instance: RuntimeInstanceIntStr):
        instance = instance.name if isinstance(instance, RuntimeInstance) else instance
        if isinstance(instance, int):
            super(NonExistingInstanceError, self).__init__('Instance with id "{}" does not exist'.format(instance))
        else:
            super(NonExistingInstanceError, self).__init__('Instance with name "{}" does not exist'.format(instance))


class TaskNotInProjectError(MetadataError):
    def __init__(self, task: Task):
        super(TaskNotInProjectError, self).__init__("Can't save task {} without project".format(task.name))

class ExperimentNotInTaskError(MetadataError):
    def __init__(self, experiment: Experiment):
        super(ExperimentNotInTaskError, self).__init__("Can't save experiment {} without task".format(experiment.name))

class ModelNotInTaskError(MetadataError):
    def __init__(self, model: Model):
        super(ModelNotInTaskError, self).__init__("Can't save model {} without task".format(model.name))


class PipelineNotInTaskError(MetadataError):
    def __init__(self, pipeline: Pipeline):
        super(PipelineNotInTaskError, self).__init__("Can't save pipeline {} without task".format(pipeline.name))


class ImageNotInTaskError(MetadataError):
    def __init__(self, image: Image):
        super(ImageNotInTaskError, self).__init__("Can't save image {} without task".format(image.name))


class InstanceNotInImageError(MetadataError):
    def __init__(self, instance: RuntimeInstance):
        super(InstanceNotInImageError, self).__init__(f"Can't save instance {instance.name} without image")


class InstanceNotInEnvironmentError(MetadataError):
    def __init__(self, instance: RuntimeInstance):
        super(InstanceNotInEnvironmentError, self).__init__(f"Can't save instance {instance.name} without environment")


class ModelWithoutIdError(MetadataError):
    def __init__(self, model: ModelIntStr):
        model = model.name if isinstance(model, Model) else model
        super(ModelWithoutIdError, self).__init__('Model "{}" id is None'.format(model))


class UnboundObjectError(MetadataError):
    pass


class ProjectWithTasksError(MetadataError):
    def __init__(self, project: Project):
        project = project.name if isinstance(project, Project) else project
        super(ProjectWithTasksError, self).__init__(f'Project {project} has foreign key and can not be deleted')


class TaskWithFKError(MetadataError):
    def __init__(self, task: Task):
        task = task.name if isinstance(task, Project) else task
        super(TaskWithFKError, self).__init__(f'Task {task} has foreign key and can not be deleted')


class ImageWithInstancesError(MetadataError):
    def __init__(self, image: Image):
        image = image.name if isinstance(image, Image) else image
        super(ImageWithInstancesError, self).__init__(f'Image {image} has foreign key and can not be deleted')


class EnvironmentWithInstancesError(MetadataError):
    def __init__(self, environment: RuntimeEnvironment):
        environment = environment.name if isinstance(environment, RuntimeEnvironment) else environment
        super(EnvironmentWithInstancesError, self).__init__(f'Environment {environment} '
                                                            f'has foreign key and can not be deleted')


class UnknownMetadataError(MetadataError):
    def __init__(self):
        super(UnknownMetadataError, self).__init__('Unknown error during query execution')


class DatasetError(AlgoLinkError):
    """
    Base class for exceptions in :class:`~algolink.repository.dataset.DatasetRpository`
    """


class NoSuchDataset(DatasetError):
    def __init__(self, dataset_id, repo, e=None):
        super(NoSuchDataset, self).__init__(f'No dataset with id {dataset_id} found in {repo}', e)


class DatasetExistsError(DatasetError):
    def __init__(self, dataset_id, repo, e=None):
        super(DatasetExistsError, self).__init__(f'Dataset with id {dataset_id} already in {repo}', e)


class ArtifactError(AlgoLinkError):
    """
    Base class for exceptions in :class:`ArtifactRepository`
    """


class NoSuchArtifactError(ArtifactError):
    """
    Exception which is thrown if artifact is not found in the repository
    """

    def __init__(self, artifact_id, repo):
        super(NoSuchArtifactError, self).__init__(f'No artifact with id {artifact_id} found in {repo}')


class ArtifactExistsError(ArtifactError):
    """
    Exception which is thrown if artifact already exists in the repository
    """

    def __init__(self, artifact_id, repo):
        super(ArtifactExistsError, self).__init__(f'Artifact with id {artifact_id} already in {repo}')



from abc import abstractmethod
from typing import Any, Dict, Iterable, List, Optional, Type, TypeVar

from pyjackson import dumps, loads
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, Text, UniqueConstraint
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from algolink.core.objects import DatasetType
from algolink.core.objects.artifacts import ArtifactCollection
from algolink.core.objects.core import (Buildable, EvaluationResults, EvaluationSet, Image, Model, Pipeline,
                                       PipelineStep, Project, RuntimeEnvironment, RuntimeInstance, Task)
from algolink.core.objects.dataset_source import DatasetSource
from algolink.core.objects.metric import Metric
from algolink.core.objects.requirements import Requirements

SQL_OBJECT_FIELD = '_sqlalchemy_object'


def json_column():
    return Column(Text)


def safe_loads(payload, as_class):
    return loads(payload, Optional[as_class])


def sqlobject(obj):
    return getattr(obj, SQL_OBJECT_FIELD, None)


def update_attrs(obj, **attrs):
    for name, value in attrs.items():
        setattr(obj, name, value)


T = TypeVar('T')
S = TypeVar('S', bound='Attaching')


class Attaching:
    id = ...
    name = ...

    def attach(self, obj):
        setattr(obj, SQL_OBJECT_FIELD, self)
        return obj

    @classmethod
    def from_obj(cls: Type[S], obj: T, new=False) -> S:
        kwargs = cls.get_kwargs(obj)
        existing = sqlobject(obj)
        if not new and existing is not None:
            update_attrs(existing, **kwargs)
            return existing
        return cls(**kwargs)

    @classmethod
    @abstractmethod
    def get_kwargs(cls, obj: T) -> dict:
        pass  # pragma: no cover

    @abstractmethod
    def to_obj(self) -> T:
        pass  # pragma: no cover


Base = declarative_base()


class SProject(Base, Attaching):
    __tablename__ = 'projects'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=True, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)

    tasks: Iterable['STask'] = relationship("STask", back_populates="project")

    def to_obj(self) -> Project:
        p = Project(self.name, id=self.id, author=self.author, creation_date=self.creation_date)
        for task in self.tasks:
            p._tasks.add(task.to_obj())
        return self.attach(p)

    @classmethod
    def get_kwargs(cls, project: Project) -> dict:
        return dict(id=project.id,
                    name=project.name,
                    author=project.author,
                    creation_date=project.creation_date,
                    tasks=[STask.from_obj(t) for t in project.tasks.values()])


class STask(Base, Attaching):
    __tablename__ = 'tasks'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)
    project_id = Column(Integer, ForeignKey('projects.id'), nullable=False)

    project = relationship("SProject", back_populates="tasks")
    models: Iterable['SModel'] = relationship("SModel", back_populates="task")
    pipelines: Iterable['SPipeline'] = relationship("SPipeline", back_populates='task')
    images: Iterable['SImage'] = relationship("SImage", back_populates='task')

    datasets = Column(Text)
    metrics = Column(Text)
    evaluation_sets = Column(Text)

    __table_args__ = (UniqueConstraint('name', 'project_id', name='tasks_name_and_ref'),)

    def to_obj(self) -> Task:
        task = Task(id=self.id,
                    name=self.name,
                    author=self.author,
                    creation_date=self.creation_date,
                    project_id=self.project_id,
                    datasets=safe_loads(self.datasets, Dict[str, DatasetSource]),
                    metrics=safe_loads(self.metrics, Dict[str, Metric]),
                    evaluation_sets=safe_loads(self.evaluation_sets, Dict[str, EvaluationSet]))
        for model in self.models:
            task._models.add(model.to_obj())

        for pipeline in self.pipelines:
            task._pipelines.add(pipeline.to_obj())

        for image in self.images:
            task._images.add(image.to_obj())
        return self.attach(task)

    @classmethod
    def get_kwargs(cls, task: Task) -> dict:
        return dict(id=task.id,
                    name=task.name,
                    author=task.author,
                    creation_date=task.creation_date,
                    project_id=task.project_id,
                    models=[SModel.from_obj(m) for m in task.models.values()],
                    images=[SImage.from_obj(i) for i in task.images.values()],
                    pipelines=[SPipeline.from_obj(p) for p in task.pipelines.values()],
                    datasets=dumps(task.datasets),
                    metrics=dumps(task.metrics),
                    evaluation_sets=dumps(task.evaluation_sets))


class SModel(Base, Attaching):
    __tablename__ = 'models'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)
    wrapper = Column(Text)

    artifact = Column(Text)
    requirements = Column(Text)
    description = Column(Text)
    params = Column(Text)
    task_id = Column(Integer, ForeignKey('tasks.id'), nullable=False)
    task = relationship("STask", back_populates="models")

    evaluations = Column(Text)
    __table_args__ = (UniqueConstraint('name', 'task_id', name='models_name_and_ref'),)

    def to_obj(self) -> Model:
        model = Model(name=self.name,
                      wrapper_meta=safe_loads(self.wrapper, dict),
                      author=self.author,
                      creation_date=self.creation_date,
                      artifact=safe_loads(self.artifact, ArtifactCollection),
                      requirements=safe_loads(self.requirements, Requirements),
                      description=self.description,
                      params=safe_loads(self.params, Dict[str, Any]),
                      id=self.id,
                      task_id=self.task_id,
                      evaluations=safe_loads(self.evaluations, Dict[str, EvaluationResults]))
        return self.attach(model)

    @classmethod
    def get_kwargs(cls, model: Model) -> dict:
        return dict(id=model.id,
                    name=model.name,
                    author=model.author,
                    creation_date=model.creation_date,
                    wrapper=dumps(model.wrapper_meta),
                    artifact=dumps(model.artifact),
                    requirements=dumps(model.requirements),
                    description=model.description,
                    params=dumps(model.params),
                    task_id=model.task_id,
                    evaluations=dumps(model.evaluations))


class SPipeline(Base, Attaching):
    __tablename__ = 'pipelines'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)
    steps = Column(Text)

    input_data = Column(Text)
    output_data = Column(Text)

    task_id = Column(Integer, ForeignKey('tasks.id'), nullable=False)
    task = relationship("STask", back_populates="pipelines")

    evaluations = Column(Text)
    __table_args__ = (UniqueConstraint('name', 'task_id', name='pipelines_name_and_ref'),)

    def to_obj(self) -> Pipeline:
        pipeline = Pipeline(name=self.name,
                            steps=safe_loads(self.steps, List[PipelineStep]),
                            input_data=safe_loads(self.input_data, DatasetType),
                            output_data=safe_loads(self.output_data, DatasetType),
                            author=self.author,
                            creation_date=self.creation_date,
                            id=self.id,
                            task_id=self.task_id,
                            evaluations=safe_loads(self.evaluations, EvaluationResults))
        return self.attach(pipeline)

    @classmethod
    def get_kwargs(cls, pipeline: Pipeline) -> dict:
        return dict(id=pipeline.id,
                    name=pipeline.name,
                    author=pipeline.author,
                    creation_date=pipeline.creation_date,
                    steps=dumps(pipeline.steps),
                    input_data=dumps(pipeline.input_data),
                    output_data=dumps(pipeline.output_data),
                    task_id=pipeline.task_id,
                    evaluations=dumps(pipeline.evaluations))


class SImage(Base, Attaching):
    __tablename__ = 'images'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)

    task_id = Column(Integer, ForeignKey('tasks.id'), nullable=False)
    task = relationship("STask", back_populates="images")

    environment_id = Column(Integer, ForeignKey('environments.id'), nullable=False)

    params = Column(Text)
    source = Column(Text)

    __table_args__ = (UniqueConstraint('name', 'task_id', name='image_name_and_ref'),)

    def to_obj(self) -> Image:
        image = Image(name=self.name,
                      author=self.author,
                      creation_date=self.creation_date,
                      id=self.id,
                      task_id=self.task_id,
                      params=safe_loads(self.params, Image.Params),
                      source=safe_loads(self.source, Buildable),
                      environment_id=self.environment_id)
        return self.attach(image)

    @classmethod
    def get_kwargs(cls, image: Image) -> dict:
        return dict(id=image.id,
                    name=image.name,
                    author=image.author,
                    creation_date=image.creation_date,
                    task_id=image.task_id,
                    params=dumps(image.params),
                    source=dumps(image.source),
                    environment_id=image.environment_id)


class SRuntimeEnvironment(Base, Attaching):
    __tablename__ = 'environments'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=True, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)

    params = Column(Text)

    def to_obj(self) -> RuntimeEnvironment:
        environment = RuntimeEnvironment(
            name=self.name,
            author=self.author,
            creation_date=self.creation_date,
            id=self.id,
            params=safe_loads(self.params, RuntimeEnvironment.Params))
        return self.attach(environment)

    @classmethod
    def get_kwargs(cls, environment: RuntimeEnvironment) -> dict:
        return dict(id=environment.id,
                    name=environment.name,
                    author=environment.author,
                    creation_date=environment.creation_date,
                    params=dumps(environment.params))


class SRuntimeInstance(Base, Attaching):
    __tablename__ = 'instances'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)

    image_id = Column(Integer, ForeignKey('images.id'), nullable=False)
    environment_id = Column(Integer, ForeignKey('environments.id'), nullable=False)

    params = Column(Text)

    __table_args__ = (UniqueConstraint('name', 'image_id', 'environment_id', name='instance_name_and_ref'),)

    def to_obj(self) -> RuntimeInstance:
        instance = RuntimeInstance(
            name=self.name,
            author=self.author,
            creation_date=self.creation_date,
            id=self.id,
            image_id=self.image_id,
            environment_id=self.environment_id,
            params=safe_loads(self.params, RuntimeInstance.Params))
        return self.attach(instance)

    @classmethod
    def get_kwargs(cls, instance: RuntimeInstance) -> dict:
        return dict(id=instance.id,
                    name=instance.name,
                    author=instance.author,
                    creation_date=instance.creation_date,
                    image_id=instance.image_id,
                    environment_id=instance.environment_id,
                    params=dumps(instance.params))



from typing import Generic, TypeVar

T = TypeVar('T')


class IndexDict(dict, Generic[T]):
    def __init__(self, key_field, index_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.key_field = key_field
        self.index_field = index_field
        self.index = {}

    def __setitem__(self, key, value):
        if key != getattr(value, self.key_field):
            raise KeyError(f'Key {key} is not equal to {self.key_field} field of {value}')
        old_value = self.get(key, None)
        if old_value is not None:
            del self.index[getattr(old_value, self.index_field)]
        super(IndexDict, self).__setitem__(key, value)
        self.index[getattr(value, self.index_field)] = key

    def add(self, value: T):
        self[getattr(value, self.key_field)] = value

    def get_index(self, key, default=...) -> T:
        if default is ...:
            return self[self.index[key]]
        else:
            key = self.index.get(key, None)
            if key is None:
                return default
            return self[key]

    def reindex(self):
        self.index = {getattr(v, self.index_field): k for k, v in self.items()}

    def clear(self) -> None:
        super(IndexDict, self).clear()
        self.index.clear()

    def __delitem__(self, key):
        value = self[key]
        del self.index[getattr(value, self.index_field)]
        super(IndexDict, self).__delitem__(key)

    def __repr__(self):
        return f'<IndexDict>{{{", ".join(f"{getattr(v, self.index_field)}[{k}]={v}" for k, v in self.items())}}}'

    def __str__(self):
        return self.__repr__()


class IndexDictAccessor(Generic[T]):
    def __init__(self, data: IndexDict[T]):
        self.data = data

    def __getitem__(self, item) -> T:
        return self.data[item]

    def __call__(self, key, default=...) -> T:
        return self.data.get_index(key, default)

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def __contains__(self, item):
        return item in self.data

    def contains(self, item):
        return item in self.data.index

    def values(self):
        return self.data.values()

    def keys(self):
        return self.data.keys()

    def items(self):
        return self.data.items()

    def get(self, key, default=...) -> T:
        return self.data.get(key, default)

    def __repr__(self):
        return f'<Accessor>{self.data}'

    def __str__(self):
        return self.__repr__()



"""
Created on Oct 30, 2021
@author: leepand6@gmail.com
"""

from math import ceil
from sqlalchemy import func

from .exceptions import PageNotAnInteger, EmptyPage


class Paginator(object):
    """
    This class helps you to manage data with pagination. This class will fetch
    data in pages. That is, instead of fetching all the records from database
    at a time, this class will fetch defined number of records at a time.
    
    So that you can perform particular action(s) on fetched data and then fetch
    data for next page. In this way we can get ride from the memory overloading
    problem as well.
    
    This class will also optimized the query for fetching total number of 
    records from database against given query_set. Optimization will be applied
    only on the query that will be used for fetching total number of records.
    You can also provide the separate query in optional_count_query_set argument
    for fetching total number of records.
    
    ..usage::
        You can use this paginator module in python scripting code and in web 
        based application code as well.
        
        :Example1:
        >>> from sqlalchemy_paginator import Paginator
        >>> query = session.query(MyModel)
        >>> paginator = Paginator(query, 5)
        >>> for page in paginator:
        >>>     print page.number  # page number of current page in iterator
        >>>     print page.object_list  # this is a list that contains the records of current page
        
        :Example2:
        >>> from sqlalchemy_paginator import Paginator
        >>> query = session.query(MyModel)
        >>> paginator = Paginator(query, 5)
        >>> page = paginator.page(page_number)
        >>> print page.paginator.count  # to get total number of records against given query
        >>> print page.paginator.total_pages  # to get total number of pages
        >>> print page.paginator.pages_range  # to get range of pages in list
        >>> print page.start_index  # to get index of the first object on this page
        >>> print page.end_index  # to get index of the last object on this page
        >>> if page.has_previous():
        >>>     print page.previous_page_number  # to get previous page number
        >>> if page.has_next():
        >>>     print page.next_page_number
    """
    def __init__(self, query_set, per_page_limit, optional_count_query_set=None,
                 allow_empty_first_page=True):
        """
        Constructor to create the paginator object.
        
        :param query_set: SQLAlchemy query which is used for fetching data from
                          from database.
        :type query_set: SQLAlchemy query object. 
        :param per_page_limit: Required number of records in a page.
        :type per_page_limit: int.
        :param optional_count_query_set: This is a optional query set that will
                                         use to fetch the total number of
                                         records from database. If this optional
                                         query is not provided than this class
                                         will optimized query_set query and used
                                         that optimized query of getting total
                                         number of records from database.
        :type optional_count_query_set: SQLAlchemy query object.
        :param allow_empty_first_page: If this flag is true and there is no
                                       data in database against given query then
                                       it will return empty list on getting
                                       first page otherwise this will raise
                                       EmptyPage error. Default value of this
                                       parameter is true.
        :type allow_empty_first_page: bool.
        """
        self.query_set = query_set
        self.per_page_limit = per_page_limit
        self.optional_count_query_set = optional_count_query_set
        self.allow_empty_first_page = allow_empty_first_page
        self.__total_pages = self.__count = None
        self.__iter_page = 1
        
    def __iter__(self):
        """The __iter__ returns the iterator object and is implicitly called at
        the start of loops"""
        self.__iter_page = 1
        return self
    
    def __next__(self):
        """Returns the next page and is implicitly called at each loop 
        increment."""
        if self.__iter_page > self.total_pages:
            raise StopIteration
        page = self.page(self.__iter_page)
        self.__iter_page += 1
        return page
    
    next = __next__

    def validate_page_number(self, page_number):
        """
        This method valid that if given page number is valid or not. Like page
        number should be integer and greater than zero and should not be greater
        than total number of pages.
        
        :param page_number: Required page number against which you want to fetch
                            records from database.
        :type page_number: int.
        :return: If given page number is valid then return it.
        :rtype: int.
        
        ..warning::
            This function can raise the following exceptions
            - PageNotAnInteger
            - EmptyPage
        """
        try:
            page_number = int(page_number)
        except ValueError:
            raise PageNotAnInteger('That page number is not an integer')
        if page_number < 1:
            raise EmptyPage('That page number is less than 1')
        if page_number > self.total_pages:
            if page_number == 1 and self.allow_empty_first_page:
                pass
            else:
                raise EmptyPage('That page contains no results')
        return page_number

    def page(self, page_number):
        """
        Returns a page object against given page number if given page number is
        valid.
        
        :param page_number: Required page number against which you want to fetch
                            records from database.
        :type page_number: int.
        :return: Page object that contains the records against given page 
                 number.
        :rtype: Page.
        
        ..seealso::
            - Page class
            - Paginator.validate_page_number()
        
        ..warning::
            This function can raise the following exceptions
            - PageNotAnInteger
            - EmptyPage
        """
        page_number = self.validate_page_number(page_number)
        offset = (page_number - 1) * self.per_page_limit
        return Page(self.query_set.offset(offset).limit(self.per_page_limit).all(),
                    page_number, self)

    def __get_count(self):
        """
        Returns the total number of objects, across all pages.
        
        :return: Total number of records against given query.
        :rtype: int.
        
        ..info::
            If optional_count_query_set is given then this function will use
            query for fetching total number records otherwise query_set query
            will be used for fetching total number records.
        """
        if self.__count is None:
            if self.optional_count_query_set is None:
                self.optional_count_query_set = self.query_set.order_by(None)
            count_query = self.optional_count_query_set.statement.with_only_columns([func.count()])
            self.__count = self.optional_count_query_set.session.execute(count_query).scalar()
        return self.__count
    count = property(__get_count)

    def __get_total_pages(self):
        """
        Returns the total number of pages.
        
        :return: Total number of pages against given query.
        :rtype: int.
        
        ..info::
            If total number of records is zero and allow_empty_first_page is
            true then returns 1 instead of 0.
        """
        ""
        if self.__total_pages is None:
            if self.count == 0 and not self.allow_empty_first_page:
                self.__total_pages = 0
            else:
                hits = max(1, self.count)
                self.__total_pages = int(ceil(hits / float(self.per_page_limit)))
        return self.__total_pages
    total_pages = property(__get_total_pages)

    def __pages_range(self):
        """
        Returns a range of pages.
        
        :return: List that contains range of pages.
        :rtype: list.
        """
        return range(1, self.total_pages + 1)
    pages_range = property(__pages_range)


class Page(object):
    """
    This is a same copy of django Page class in paginator module. This class
    will be used in Paginator class for making pages. This Page class contains
    a list of objects of one page, page number and reference of paginator
    instance.
    """
    def __init__(self, object_list, number, paginator):
        self.object_list = object_list
        self.number = number
        self.paginator = paginator

    def __repr__(self):
        return '<Page %s of %s>' % (self.number, self.paginator.total_pages)

    def has_next(self):
        return self.number < self.paginator.total_pages

    def has_previous(self):
        return self.number > 1

    def has_other_pages(self):
        return self.has_previous() or self.has_next()

    def __next_page_number(self):
        return self.number + 1
    next_page_number = property(__next_page_number)

    def __previous_page_number(self):
        return self.number - 1
    previous_page_number = property(__previous_page_number)

    def __start_index(self):
        """
        Returns the index of the first object on this page,
        relative to total objects in the paginator.
        """
        # Special case, return zero if no items.
        if self.paginator.count == 0:
            return 0
        return (self.paginator.per_page_limit * (self.number - 1)) + 1
    start_index = property(__start_index)

    def __end_index(self):
        """
        Returns the index of the last object on this page,
        relative to total objects found (hits).
        """
        # Special case for the last page
        if self.number == self.paginator.total_pages:
            return self.paginator.count
        return self.number * self.paginator.per_page_limit
    end_index = property(__end_index)



import os
import shutil
from abc import abstractmethod
from contextlib import contextmanager

from algolink.build.provider import PythonProvider
from algolink.core.objects import core
from algolink.utils.fs import get_lib_path
from algolink.utils.log import logger

REQUIREMENTS = 'requirements.txt'
_ALGOLINK_SOURCE = True
# _ALGOLINK_SOURCE defines in which way algolink will be installed inside of the instance. Depending on it's value
# True - means that it will install algolink from PIP
# False - That it will use local algolink installation
# str, representing a path - that it will search for .whl file with algolink package


def algolink_from_pip():
    """
    :return boolen flag if algolink inside image must be installed from pip (or copied local dist instread)"""
    return _ALGOLINK_SOURCE


@contextmanager
def use_local_installation():
    """Context manager that changes docker builder behaviour to copy
    this installation of algolink instead of installing it from pip.
    This is needed for testing and examples"""
    global _ALGOLINK_SOURCE
    tmp = _ALGOLINK_SOURCE
    _ALGOLINK_SOURCE = False
    try:
        yield
    finally:
        _ALGOLINK_SOURCE = tmp


@contextmanager
def use_wheel_installation(path: str):
    """Context manager that changes docker builder behaviour to
    install algolink from wheel.
    This is needed in the case you using algolink from wheel"""
    global _ALGOLINK_SOURCE
    tmp = _ALGOLINK_SOURCE
    _ALGOLINK_SOURCE = path
    try:
        yield
    finally:
        _ALGOLINK_SOURCE = tmp


class BuilderBase:
    """Abstract class for building images from algolink objects"""

    @abstractmethod
    def create_image(self, name: str, environment: 'core.RuntimeEnvironment', **kwargs) -> 'core.Image.Params':
        """Abstract method to create image"""

    @abstractmethod
    def build_image(self, buildable: 'core.Buildable', image: 'core.Image.Params',
                    environment: 'core.RuntimeEnvironment.Params', **kwargs):
        """Abstract method to build image"""

    @abstractmethod
    def delete_image(self, image: 'core.Image.Params', environment: 'core.RuntimeEnvironment.Params', **kwargs):
        """Abstract method to delete image"""

    @abstractmethod
    def image_exists(self, image: 'core.Image.Params', environment: 'core.RuntimeEnvironment.Params', **kwargs):
        """Abstract method to check if image exists"""


class PythonBuildContext:
    """
    Basic class for building python images from algolink objects

    :param provider: A ProviderBase instance to get distribution from
    """

    def __init__(self, provider: PythonProvider):
        self.provider = provider

    def _write_distribution(self, target_dir):
        """
        Writes full distribution to dir
        :param target_dir: target directory to write distribution
        """
        logger.debug('Writing model distribution to "%s"...', target_dir)
        self._write_sources(target_dir)
        self._write_binaries(target_dir)
        self._write_requirements(target_dir)
        self._write_run_script(target_dir)

    def _write_sources(self, target_dir):
        """
        Writes sources to dir
        :param target_dir: target directory to write sources
        """
        for name, content in self.provider.get_sources().items():
            logger.debug('Putting model source "%s" to distribution...', name)
            path = os.path.join(target_dir, name)
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, 'w' if isinstance(content, str) else 'wb', encoding='utf8') as src:
                src.write(content)

        pip_algolink = algolink_from_pip()
        if pip_algolink is False:
            logger.debug('Putting AlgoLink sources to distribution as local installation is employed...')
            main_module_path = get_lib_path('.')
            shutil.copytree(main_module_path, os.path.join(target_dir, 'algolink'))
        elif isinstance(pip_algolink, str):
            logger.debug('Putting AlgoLink wheel to distribution as wheel installation is employed...')
            shutil.copy(pip_algolink, target_dir)

    def _write_binaries(self, path):
        """
        Writes binaries to dir
        :param path: target directory to write binaries
        """
        logger.debug('Putting model artifacts to distribution...')
        a = self.provider.get_artifacts()
        a.materialize(path)

    def _write_requirements(self, target_dir):
        """
        Writes requirements.txt to dir
        :param target_dir: target directory to write requirements
        """
        with open(os.path.join(target_dir, REQUIREMENTS), 'w', encoding='utf8') as req:
            requirements = self.provider.get_requirements()
            logger.debug('Auto-determined requirements for model: %s.', requirements.to_pip())
            if algolink_from_pip() is False:
                cwd = os.getcwd()
                try:
                    from setup import setup_args  # FIXME only for development
                    requirements += list(setup_args['install_requires'])
                    logger.debug('Adding AlgoLink requirements as local installation is employed...')
                    logger.debug('Overall requirements for model: %s.', requirements.to_pip())
                finally:
                    os.chdir(cwd)
            req.write('\n'.join(requirements.to_pip()))

    def _write_run_script(self, target_dir):
        """
        Writes run.sh script to dir
        :param target_dir: target directory to script
        """
        with open(os.path.join(target_dir, 'run.sh'), 'w') as sh:
            sh.write('python -c "from algolink import start_runtime; start_runtime()"')
